
# __init__.py
# Entry point for anaconda storage formats subpackage.
#
# Copyright (C) 2009  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Dave Lehman <dlehman@redhat.com>
#

import gi
gi.require_version("BlockDev", "2.0")

from gi.repository import BlockDev as blockdev

import os
import importlib
from six import add_metaclass

from .. import udev
from ..util import get_sysfs_path_by_name
from ..util import run_program
from ..util import ObjectID
from ..storage_log import log_method_call
from ..errors import DeviceFormatError, FormatCreateError, FormatDestroyError, FormatSetupError
from ..i18n import N_
from ..size import Size, ROUND_DOWN, unit_str
from ..threads import SynchronizedMeta
from ..flags import flags

from ..errors import FSError, FSResizeError, LUKSError, FormatResizeError

from ..tasks import fsinfo
from ..tasks import fsresize
from ..tasks import fssize
from ..tasks import fsck
from ..tasks import fsminsize

import logging
log = logging.getLogger("blivet")

device_formats = {}


def register_device_format(fmt_class):
    if not issubclass(fmt_class, DeviceFormat):
        raise ValueError("arg1 must be a subclass of DeviceFormat")

    device_formats[fmt_class._type] = fmt_class
    log.debug("registered device format class %s as %s", fmt_class.__name__,
              fmt_class._type)


default_fstypes = ("ext4", "xfs", "ext3", "ext2")


def get_default_filesystem_type():
    for fstype in default_fstypes:
        try:
            supported = get_format(fstype).supported
        except AttributeError:
            supported = False

        if supported:
            return fstype

    raise DeviceFormatError("None of %s is supported by your kernel" % ",".join(default_fstypes))


def get_format(fmt_type, *args, **kwargs):
    """ Return an instance of the appropriate DeviceFormat class.

        :param fmt_type: The name of the formatting type
        :type fmt_type: str.
        :return: the format instance
        :rtype: :class:`DeviceFormat`
        :raises: ValueError

        .. note::

            Any additional arguments will be passed on to the constructor for
            the format class. See the various :class:`DeviceFormat` subclasses
            for an exhaustive list of the arguments that can be passed.
    """
    fmt_class = get_device_format_class(fmt_type)
    if not fmt_class:
        fmt_class = DeviceFormat
    fmt = fmt_class(*args, **kwargs)

    # this allows us to store the given type for formats we implement as
    # DeviceFormat.
    if fmt_type and fmt.type is None:
        # unknown type, but we can set the name of the format
        # this should add/set an instance attribute
        fmt._name = fmt_type

    log.debug("get_format('%s') returning %s instance with object id %d",
              fmt_type, fmt.__class__.__name__, fmt.id)
    return fmt


def get_device_format_class(fmt_type):
    """ Return an appropriate format class.

        :param fmt_type: The name of the format type.
        :type fmt_type: str.
        :returns: The chosen DeviceFormat class
        :rtype: class.

        Returns None if no class is found for fmt_type.
    """
    fmt = device_formats.get(fmt_type)
    if not fmt:
        for fmt_class in device_formats.values():
            if fmt_type and fmt_type == fmt_class._name:
                fmt = fmt_class
                break
            elif fmt_type in fmt_class._udev_types:
                fmt = fmt_class
                break

    return fmt


@add_metaclass(SynchronizedMeta)
class DeviceFormat(ObjectID):

    """ Generic device format.

        This represents the absence of recognized formatting. That could mean a
        device is uninitialized, has had zeros written to it, or contains some
        valid formatting that this module does not support.
    """
    _type = None
    _name = N_("Unknown")
    _udev_types = []
    parted_flag = None
    parted_system = None
    _formattable = False                # can be formatted
    _supported = False                  # is supported
    _linux_native = False                # for clearpart
    _packages = []                      # required packages
    _resizable = False                  # can be resized
    _max_size = Size(0)                  # maximum size
    _min_size = Size(0)                  # minimum size
    _dump = False
    _check = False
    _hidden = False                     # hide devices with this formatting?
    _ks_mountpoint = None

    _resize_class = fsresize.UnimplementedFSResize
    _size_info_class = fssize.UnimplementedFSSize
    _info_class = fsinfo.UnimplementedFSInfo
    _minsize_class = fsminsize.UnimplementedFSMinSize

    def __init__(self, **kwargs):
        """
            :keyword device: The path to the device node.
            :type device: str
            :keyword uuid: the formatting's UUID.
            :type uuid: str
            :keyword exists: Whether the formatting exists. (default: False)
            :raises: ValueError

            .. note::

                The 'device' kwarg is required for existing formats. For non-
                existent formats, it is only necessary that the :attr:`device`
                attribute be set before the :meth:`create` method runs. Note
                that you can specify the device at the last moment by specifying
                it via the 'device' kwarg to the :meth:`create` method.
        """
        ObjectID.__init__(self)
        self._label = None
        self._options = None
        self._device = None

        self.device = kwargs.get("device")
        self.uuid = kwargs.get("uuid")
        self.exists = kwargs.get("exists", False)
        self.options = kwargs.get("options")
        self._create_options = kwargs.get("create_options")

        # Create task objects
        self._info = self._info_class(self)
        self._resize = self._resize_class(self)
        # These two may depend on info class, so create them after
        self._minsize = self._minsize_class(self)
        self._size_info = self._size_info_class(self)

        # format size does not necessarily equal device size
        self._size = kwargs.get("size", Size(0))
        self._target_size = self._size
        self._min_instance_size = Size(0)    # min size of this DeviceFormat instance

        # Resize operations are limited to error-free formats whose current
        # size is known.
        self._resizable = False

    def __repr__(self):
        s = ("%(classname)s instance (%(id)s) object id %(object_id)d--\n"
             "  type = %(type)s  name = %(name)s  status = %(status)s\n"
             "  device = %(device)s  uuid = %(uuid)s  exists = %(exists)s\n"
             "  options = %(options)s\n"
             "  create_options = %(create_options)s  supported = %(supported)s"
             "  formattable = %(format)s  resizable = %(resize)s\n" %
             {"classname": self.__class__.__name__, "id": "%#x" % id(self),
              "object_id": self.id,
              "type": self.type, "name": self.name, "status": self.status,
              "device": self.device, "uuid": self.uuid, "exists": self.exists,
              "options": self.options, "supported": self.supported,
              "format": self.formattable, "resize": self.resizable,
              "create_options": self.create_options})
        return s

    @property
    def _existence_str(self):
        return "existing" if self.exists else "non-existent"

    @property
    def desc(self):
        return str(self.type)

    def __str__(self):
        return "%s %s" % (self._existence_str, self.desc)

    @property
    def dict(self):
        d = {"type": self.type, "name": self.name, "device": self.device,
             "uuid": self.uuid, "exists": self.exists,
             "options": self.options, "supported": self.supported,
             "resizable": self.resizable, "create_options": self.create_options}

        return d

    def labeling(self):
        """Returns False by default since most formats are non-labeling."""
        return False

    def relabels(self):
        """Returns False by default since most formats are non-labeling."""
        return False

    def label_format_ok(self, label):
        """Checks whether the format of the label is OK for whatever
           application is used by blivet to write a label for this format.
           If there is no application that blivet uses to write a label,
           then no format is acceptable, so must return False.

           :param str label: The label to be checked

           :rtype: bool
           :return: True if the format of the label is OK, otherwise False
        """
        # pylint: disable=unused-argument
        return self.labeling()

    def _set_label(self, label):
        """Sets the label for this format.

           :param label: the label for this format
           :type label: str or None

           None means no label specified, or in other words, accept the default
           label that the filesystem app may set. Once the device exists the
           label should not be None, as the device must then have some label
           even if just the empty label.

           "" means the empty label, i.e., no label.

           Some filesystems, even though they do not have a
           labeling application may be already labeled, so we allow to set
           the label of a filesystem even if a labeling application does not
           exist. This can happen with the install media, for example, where
           the filesystem on the CD has a label, but there is no labeling
           application for the Iso9660FS format.

           If a labeling application does exist, the label is not
           required to have the correct format for that application.
           The allowable format for the label may be more permissive than
           the format allowed by the labeling application.

           This method is not intended to be overridden.
        """
        self._label = label

    def _get_label(self):
        """The label for this filesystem.

           :return: the label for this device
           :rtype: str

           This method is not intended to be overridden.
        """
        return self._label

    def _set_options(self, options):
        self._options = options

    def _get_options(self):
        return self._options

    options = property(
        lambda s: s._get_options(),
        lambda s, v: s._set_options(v),
        doc="fstab entry option string"
    )

    def _set_create_options(self, options):
        self._create_options = options

    def _get_create_options(self):
        return self._create_options

    create_options = property(
        lambda s: s._get_create_options(),
        lambda s, v: s._set_create_options(v),
        doc="options to be used when running mkfs"
    )

    def _device_check(self, devspec):
        """ Verifies that device spec has a proper format.

            :param devspec: the device spec
            :type devspec: str or NoneType
            :rtype: str or NoneType
            :returns: an explanatory message if devspec fails check, else None
        """
        if devspec and not devspec.startswith("/"):
            return "device must be a fully qualified path"
        return None

    def _set_device(self, devspec):
        error_msg = self._device_check(devspec)
        if error_msg:
            raise ValueError(error_msg)
        self._device = devspec

    def _get_device(self):
        return self._device

    device = property(lambda f: f._get_device(),
                      lambda f, d: f._set_device(d),
                      doc="Full path the device this format occupies")

    @property
    def name(self):
        return self._name or self.type

    @property
    def type(self):
        return self._type

    def _set_target_size(self, newsize):
        """ Set the target size for this filesystem.

            :param :class:`~.size.Size` newsize: the newsize
        """
        if not isinstance(newsize, Size):
            raise ValueError("new size must be of type Size")

        if not self.exists:
            raise DeviceFormatError("format has not been created")

        if not self.resizable:
            raise DeviceFormatError("format is not resizable")

        if newsize < self.min_size:
            raise ValueError("requested size %s must be at least minimum size %s" % (newsize, self.min_size))

        if self.max_size and newsize >= self.max_size:
            raise ValueError("requested size %s must be less than maximum size %s" % (newsize, self.max_size))

        self._target_size = newsize

    def _get_target_size(self):
        """ Get this filesystem's target size. """
        return self._target_size

    target_size = property(_get_target_size, _set_target_size,
                           doc="Target size for this filesystem")

    def _get_size(self):
        """ Get this filesystem's size. """
        return self.target_size if self.resizable else self._size

    size = property(_get_size, doc="This filesystem's size, accounting "
                    "for pending changes")

    @property
    def min_size(self):
        # If self._min_instance_size is not 0, then it should be no less than
        # self._min_size, by definition, and since a non-zero value indicates
        # that it was obtained, it is the preferred value.
        # If self._min_instance_size is less than self._min_size,
        # but not 0, then there must be some mistake, so better to use
        # self._min_size.
        return max(self._min_instance_size, self._min_size)

    def update_size_info(self):
        """ Update this format's current and minimum size (for resize). """

    def do_resize(self):
        """ Resize this filesystem based on this instance's target_size attr.

            :raises: FSResizeError, FormatResizeError
        """
        if not self.exists:
            raise FormatResizeError("format does not exist", self.device)

        if not self.resizable:
            raise FormatResizeError("format not resizable", self.device)

        if self.target_size == self.current_size:
            return

        if not self._resize.available:
            return

        # tmpfs mounts don't need an existing device node
        if not self.device == "tmpfs" and not os.path.exists(self.device):
            raise FormatResizeError("device does not exist", self.device)

        self._pre_resize()

        self.update_size_info()

        # Check again if resizable is True, as update_size_info() can change that
        if not self.resizable:
            raise FormatResizeError("format not resizable", self.device)

        if self.target_size < self.min_size:
            self.target_size = self.min_size
            log.info("Minimum size changed, setting target_size on %s to %s",
                     self.device, self.target_size)

        # Bump target size to nearest whole number of the resize tool's units.
        # We always round down because the fs has to fit on whatever device
        # contains it. To round up would risk quietly setting a target size too
        # large for the device to hold.
        rounded = self.target_size.round_to_nearest(self._resize.unit,
                                                    rounding=ROUND_DOWN)

        # 1. target size was between the min size and max size values prior to
        #    rounding (see _set_target_size)
        # 2. we've just rounded the target size down (or not at all)
        # 3. the minimum size is already either rounded (see _get_min_size) or is
        #    equal to the current size (see update_size_info)
        # 5. the minimum size is less than or equal to the current size (see
        #    _get_min_size)
        #
        # This, I think, is sufficient to guarantee that the rounded target size
        # is greater than or equal to the minimum size.

        # It is possible that rounding down a target size greater than the
        # current size would move it below the current size, thus changing the
        # direction of the resize. That means the target size was less than one
        # unit larger than the current size, and we should do nothing and return
        # early.
        if self.target_size > self.current_size and rounded < self.current_size:
            log.info("rounding target size down to next %s obviated resize of "
                     "filesystem on %s", unit_str(self._resize.unit), self.device)
            return
        else:
            self.target_size = rounded

        try:
            self._resize.do_task()
        except FSError as e:
            raise FSResizeError(e, self.device)
        except LUKSError as e:
            raise FormatResizeError(e, self.device)

        self._post_resize()

    def _pre_resize(self):
        """ Do whatever needs to be done before the format is resized """

    def _post_resize(self):
        # XXX must be a smarter way to do this
        self._size = self.target_size

    @property
    def current_size(self):
        """ The filesystem's current actual size. """
        return self._size if self.exists else Size(0)

    def create(self, **kwargs):
        """ Write the formatting to the specified block device.

            :keyword device: path to device node
            :type device: str.
            :raises: FormatCreateError
            :returns: None.

            .. :note::

                If a device node path is passed to this method it will overwrite
                any previously set value of this instance's "device" attribute.
        """
        log_method_call(self, device=self.device,
                        type=self.type, status=self.status)
        self._pre_create(**kwargs)
        self._create(**kwargs)
        self._post_create(**kwargs)

    def _pre_create(self, **kwargs):
        """ Perform checks and setup prior to creating the format. """
        # allow late specification of device path
        device = kwargs.get("device")
        if device:
            self.device = device

        if not os.path.exists(self.device):
            raise FormatCreateError("invalid device specification %s" % self.device)

        if self.exists:
            raise DeviceFormatError("format already exists")

        if self.status:
            raise DeviceFormatError("device exists and is active")

    # pylint: disable=unused-argument
    def _create(self, **kwargs):
        """ Type-specific create method. """

    # pylint: disable=unused-argument
    def _post_create(self, **kwargs):
        self.exists = True

    def destroy(self, **kwargs):
        """ Remove the formatting from the associated block device.

            :raises: FormatDestroyError
            :returns: None.
        """
        log_method_call(self, device=self.device,
                        type=self.type, status=self.status)
        self._pre_destroy(**kwargs)
        self._destroy(**kwargs)
        self._post_destroy(**kwargs)

    # pylint: disable=unused-argument
    def _pre_destroy(self, **kwargs):
        if not self.exists:
            raise DeviceFormatError("format has not been created")

        if self.status:
            raise DeviceFormatError("device is active")

        if not os.access(self.device, os.W_OK):
            raise DeviceFormatError("device path does not exist or is not writable")

    def _destroy(self, **kwargs):
        rc = 0
        err = ""
        try:
            rc = run_program(["wipefs", "-f", "-a", self.device])
        except OSError as e:
            err = str(e)
        else:
            if rc:
                err = str(rc)

        if err:
            msg = "error wiping old signatures from %s: %s" % (self.device, err)
            raise FormatDestroyError(msg)

    def _post_destroy(self, **kwargs):
        udev.settle()
        self.exists = False

    @property
    def destroyable(self):
        """ Do we have the facilities to destroy a format of this type. """
        # assumes wipefs is always available
        return True

    def setup(self, **kwargs):
        """ Activate the formatting.

            :keyword device: device node path
            :type device: str.
            :raises: FormatSetupError.
            :returns: None.

            .. :note::

                If a device node path is passed to this method it will overwrite
                any previously set value of this instance's "device" attribute.
        """
        log_method_call(self, device=self.device,
                        type=self.type, status=self.status)
        if not self._pre_setup(**kwargs):
            return

        self._setup(**kwargs)
        self._post_setup(**kwargs)

    @property
    def controllable(self):
        """ Are external utilities available to allow this format to be both
            setup and teared down.

            :returns: True if this format can be set up, otherwise False
            :rtype: bool
        """
        return True

    def _pre_setup(self, **kwargs):
        """ Return True if setup should proceed. """
        if not self.exists:
            raise FormatSetupError("format has not been created")

        # allow late specification of device path
        device = kwargs.get("device")
        if device:
            self.device = device

        if not self.device or not os.path.exists(self.device):
            raise FormatSetupError("invalid device specification")

        return not self.status

    # pylint: disable=unused-argument
    def _setup(self, **kwargs):
        pass

    # pylint: disable=unused-argument
    def _post_setup(self, **kwargs):
        pass

    def teardown(self, **kwargs):
        """ Deactivate the formatting. """
        log_method_call(self, device=self.device,
                        type=self.type, status=self.status)
        if not self._pre_teardown(**kwargs):
            return

        self._teardown(**kwargs)
        self._post_teardown(**kwargs)

    def _pre_teardown(self, **kwargs):
        """ Return True if teardown should proceed. """
        if not self.exists:
            raise DeviceFormatError("format has not been created")

        return self.status

    def _teardown(self, **kwargs):
        pass

    def _post_teardown(self, **kwargs):
        pass

    @property
    def status(self):
        return (self.exists and
                self.__class__ is not DeviceFormat and
                isinstance(self.device, str) and
                self.device and
                os.path.exists(self.device))

    @property
    def formattable(self):
        """ Can we create formats of this type? """
        return self._formattable

    @property
    def supported(self):
        """ Is this format a supported type?

            Are the necessary external applications required by the
            functionality that this format provides actually provided by
            the environment in which blivet is running?
        """
        return self._supported

    @property
    def packages(self):
        """ Packages required to manage formats of this type. """
        return self._packages

    @property
    def resizable(self):
        """ Can formats of this type be resized? """
        return self._resizable and self.exists

    @property
    def linux_native(self):
        """ Is this format type native to linux? """
        return self._linux_native

    @property
    def mountable(self):
        """ Is this something we can mount? """
        return False

    @property
    def dump(self):
        """ Whether or not this format will be dumped by dump(8). """
        return self._dump

    @property
    def check(self):
        """ Whether or not this format is checked on boot. """
        return self._check

    @property
    def max_size(self):
        """ Maximum size for this format type. """
        return self._max_size

    @property
    def hidden(self):
        """ Whether devices with this formatting should be hidden in UIs. """
        return self._hidden

    @property
    def ks_mountpoint(self):
        return (self._ks_mountpoint or self.type or "")

    def populate_ksdata(self, data):
        data.format = not self.exists
        data.fstype = self.type
        data.mountpoint = self.ks_mountpoint


register_device_format(DeviceFormat)

# import the format modules (which register their device formats)
from . import biosboot, disklabel, dmraid, fslib, fs, luks, lvmpv, mdraid, multipath, prepboot, swap, stratis



# gpt.py
# GPT partitioning helpers
#
# Copyright (C) Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

import uuid

from blivet.arch import get_arch
from blivet.errors import GPTVolUUIDError

GPT_VOL_ARCH_ROOT = "root"
GPT_VOL_ARCH_ROOT_VERITY = "root_verity"
GPT_VOL_ARCH_ROOT_VERITY_SIG = "root_verity_sig"
GPT_VOL_ARCH_USR = "usr"
GPT_VOL_ARCH_USR_VERITY = "usr_verity"
GPT_VOL_ARCH_USR_VERITY_SIG = "usr_verity_sig"

GPT_VOL_ESP = "esp"
GPT_VOL_XBOOTLDR = "xbootldr"
GPT_VOL_SWAP = "swap"
GPT_VOL_HOME = "home"
GPT_VOL_SRV = "srv"
GPT_VOL_VAR = "var"
GPT_VOL_TMP = "tmp"
GPT_VOL_USER_HOME = "user_home"
GPT_VOL_LINUX_GENERIC = "linux_generic"

_gpt_common_uuid = {
    GPT_VOL_ESP: uuid.UUID("c12a7328-f81f-11d2-ba4b-00a0c93ec93b"),
    GPT_VOL_XBOOTLDR: uuid.UUID("bc13c2ff-59e6-4262-a352-b275fd6f7172"),
    GPT_VOL_SWAP: uuid.UUID("0657fd6d-a4ab-43c4-84e5-0933c84b4f4f"),
    GPT_VOL_HOME: uuid.UUID("933ac7e1-2eb4-4f13-b844-0e14e2aef915"),
    GPT_VOL_SRV: uuid.UUID("3b8f8425-20e0-4f3b-907f-1a25a76f98e8"),
    GPT_VOL_VAR: uuid.UUID("4d21b016-b534-45c2-a9fb-5c16e091fd2d"),
    GPT_VOL_TMP: uuid.UUID("7ec6f557-3bc5-4aca-b293-16ef5df639d1"),
    GPT_VOL_USER_HOME: uuid.UUID("773f91ef-66d4-49b5-bd83-d683bf40ad16"),
    GPT_VOL_LINUX_GENERIC: uuid.UUID("0fc63daf-8483-4772-8e79-3d69d8477de4"),
}

_gpt_arch_uuid = {
    GPT_VOL_ARCH_ROOT: {
        "alpha": uuid.UUID("6523f8ae-3eb1-4e2a-a05a-18b695ae656f"),
        "arc": uuid.UUID("d27f46ed-2919-4cb8-bd25-9531f3c16534"),
        "arm": uuid.UUID("69dad710-2ce4-4e3c-b16c-21a1d49abed3"),
        "aarch64": uuid.UUID("b921b045-1df0-41c3-af44-4c6f280d3fae"),
        "i386": uuid.UUID("44479540-f297-41b2-9af7-d131d5f0458a"),
        "ia64": uuid.UUID("993d8d3d-f80e-4225-855a-9daf8ed7ea97"),
        "loongarch64": uuid.UUID("77055800-792c-4f94-b39a-98c91b762bb6"),
        "mips64el": uuid.UUID("700bda43-7a34-4507-b179-eeb93d7a7ca3"),
        "mipsel": uuid.UUID("37c58c8a-d913-4156-a25f-48b1b64e07f0"),
        "parisc": uuid.UUID("1aacdb3b-5444-4138-bd9e-e5c2239b2346"),
        "ppc": uuid.UUID("1de3f1ef-fa98-47b5-8dcd-4a860a654d78"),
        "ppc64": uuid.UUID("912ade1d-a839-4913-8964-a10eee08fbd2"),
        "ppc64el": uuid.UUID("c31c45e6-3f39-412e-80fb-4809c4980599"),
        "riscv32": uuid.UUID("60d5a7fe-8e7d-435c-b714-3dd8162144e1"),
        "riscv64": uuid.UUID("72ec70a6-cf74-40e6-bd49-4bda08e8f224"),
        "s390": uuid.UUID("08a7acea-624c-4a20-91e8-6e0fa67d23f9"),
        "s390x": uuid.UUID("5eead9a9-fe09-4a1e-a1d7-520d00531306"),
        "tilegx": uuid.UUID("c50cdd70-3862-4cc3-90e1-809a8c93ee2c"),
        "x86_64": uuid.UUID("4f68bce3-e8cd-4db1-96e7-fbcaf984b709"),
    },
    GPT_VOL_ARCH_ROOT_VERITY: {
        "alpha": uuid.UUID("fc56d9e9-e6e5-4c06-be32-e74407ce09a5"),
        "arc": uuid.UUID("24b2d975-0f97-4521-afa1-cd531e421b8d"),
        "arm": uuid.UUID("7386cdf2-203c-47a9-a498-f2ecce45a2d6"),
        "aarch64": uuid.UUID("df3300ce-d69f-4c92-978c-9bfb0f38d820"),
        "i386": uuid.UUID("d13c5d3b-b5d1-422a-b29f-9454fdc89d76"),
        "ia64": uuid.UUID("86ed10d5-b607-45bb-8957-d350f23d0571"),
        "loongarch64": uuid.UUID("f3393b22-e9af-4613-a948-9d3bfbd0c535"),
        "mips64el": uuid.UUID("16b417f8-3e06-4f57-8dd2-9b5232f41aa6"),
        "mipsel": uuid.UUID("d7d150d2-2a04-4a33-8f12-16651205ff7b"),
        "parisc": uuid.UUID("d212a430-fbc5-49f9-a983-a7feef2b8d0e"),
        "ppc": uuid.UUID("98cfe649-1588-46dc-b2f0-add147424925"),
        "ppc64": uuid.UUID("9225a9a3-3c19-4d89-b4f6-eeff88f17631"),
        "ppc64el": uuid.UUID("906bd944-4589-4aae-a4e4-dd983917446a"),
        "riscv32": uuid.UUID("ae0253be-1167-4007-ac68-43926c14c5de"),
        "riscv64": uuid.UUID("b6ed5582-440b-4209-b8da-5ff7c419ea3d"),
        "s390": uuid.UUID("7ac63b47-b25c-463b-8df8-b4a94e6c90e1"),
        "s390x": uuid.UUID("b325bfbe-c7be-4ab8-8357-139e652d2f6b"),
        "tilegx": uuid.UUID("966061ec-28e4-4b2e-b4a5-1f0a825a1d84"),
        "x86_64": uuid.UUID("2c7357ed-ebd2-46d9-aec1-23d437ec2bf5"),
    },
    GPT_VOL_ARCH_ROOT_VERITY_SIG: {
        "alpha": uuid.UUID("d46495b7-a053-414f-80f7-700c99921ef8"),
        "arc": uuid.UUID("143a70ba-cbd3-4f06-919f-6c05683a78bc"),
        "arm": uuid.UUID("42b0455f-eb11-491d-98d3-56145ba9d037"),
        "aarch64": uuid.UUID("6db69de6-29f4-4758-a7a5-962190f00ce3"),
        "i386": uuid.UUID("5996fc05-109c-48de-808b-23fa0830b676"),
        "ia64": uuid.UUID("e98b36ee-32ba-4882-9b12-0ce14655f46a"),
        "loongarch64": uuid.UUID("5afb67eb-ecc8-4f85-ae8e-ac1e7c50e7d0"),
        "mips64el": uuid.UUID("904e58ef-5c65-4a31-9c57-6af5fc7c5de7"),
        "mipsel": uuid.UUID("c919cc1f-4456-4eff-918c-f75e94525ca5"),
        "parisc": uuid.UUID("15de6170-65d3-431c-916e-b0dcd8393f25"),
        "ppc": uuid.UUID("1b31b5aa-add9-463a-b2ed-bd467fc857e7"),
        "ppc64": uuid.UUID("f5e2c20c-45b2-4ffa-bce9-2a60737e1aaf"),
        "ppc64el": uuid.UUID("d4a236e7-e873-4c07-bf1d-bf6cf7f1c3c6"),
        "riscv32": uuid.UUID("3a112a75-8729-4380-b4cf-764d79934448"),
        "riscv64": uuid.UUID("efe0f087-ea8d-4469-821a-4c2a96a8386a"),
        "s390": uuid.UUID("3482388e-4254-435a-a241-766a065f9960"),
        "s390x": uuid.UUID("c80187a5-73a3-491a-901a-017c3fa953e9"),
        "tilegx": uuid.UUID("b3671439-97b0-4a53-90f7-2d5a8f3ad47b"),
        "x86_64": uuid.UUID("41092b05-9fc8-4523-994f-2def0408b176"),
    },
    GPT_VOL_ARCH_USR: {
        "alpha": uuid.UUID("e18cf08c-33ec-4c0d-8246-c6c6fb3da024"),
        "arc": uuid.UUID("7978a683-6316-4922-bbee-38bff5a2fecc"),
        "arm": uuid.UUID("7d0359a3-02b3-4f0a-865c-654403e70625"),
        "aarch64": uuid.UUID("b0e01050-ee5f-4390-949a-9101b17104e9"),
        "i386": uuid.UUID("75250d76-8cc6-458e-bd66-bd47cc81a812"),
        "ia64": uuid.UUID("4301d2a6-4e3b-4b2a-bb94-9e0b2c4225ea"),
        "loongarch64": uuid.UUID("e611c702-575c-4cbe-9a46-434fa0bf7e3f"),
        "mips64el": uuid.UUID("c97c1f32-ba06-40b4-9f22-236061b08aa8"),
        "mipsel": uuid.UUID("0f4868e9-9952-4706-979f-3ed3a473e947"),
        "parisc": uuid.UUID("dc4a4480-6917-4262-a4ec-db9384949f25"),
        "ppc": uuid.UUID("7d14fec5-cc71-415d-9d6c-06bf0b3c3eaf"),
        "ppc64": uuid.UUID("2c9739e2-f068-46b3-9fd0-01c5a9afbcca"),
        "ppc64el": uuid.UUID("15bb03af-77e7-4d4a-b12b-c0d084f7491c"),
        "riscv32": uuid.UUID("b933fb22-5c3f-4f91-af90-e2bb0fa50702"),
        "riscv64": uuid.UUID("beaec34b-8442-439b-a40b-984381ed097d"),
        "s390": uuid.UUID("cd0f869b-d0fb-4ca0-b141-9ea87cc78d66"),
        "s390x": uuid.UUID("8a4f5770-50aa-4ed3-874a-99b710db6fea"),
        "tilegx": uuid.UUID("55497029-c7c1-44cc-aa39-815ed1558630"),
        "x86_64": uuid.UUID("8484680c-9521-48c6-9c11-b0720656f69e"),
    },
    GPT_VOL_ARCH_USR_VERITY: {
        "alpha": uuid.UUID("8cce0d25-c0d0-4a44-bd87-46331bf1df67"),
        "arc": uuid.UUID("fca0598c-d880-4591-8c16-4eda05c7347c"),
        "arm": uuid.UUID("c215d751-7bcd-4649-be90-6627490a4c05"),
        "aarch64": uuid.UUID("6e11a4e7-fbca-4ded-b9e9-e1a512bb664e"),
        "i386": uuid.UUID("8f461b0d-14ee-4e81-9aa9-049b6fb97abd"),
        "ia64": uuid.UUID("6a491e03-3be7-4545-8e38-83320e0ea880"),
        "loongarch64": uuid.UUID("f46b2c26-59ae-48f0-9106-c50ed47f673d"),
        "mips64el": uuid.UUID("3c3d61fe-b5f3-414d-bb71-8739a694a4ef"),
        "mipsel": uuid.UUID("46b98d8d-b55c-4e8f-aab3-37fca7f80752"),
        "parisc": uuid.UUID("5843d618-ec37-48d7-9f12-cea8e08768b2"),
        "ppc": uuid.UUID("df765d00-270e-49e5-bc75-f47bb2118b09"),
        "ppc64": uuid.UUID("bdb528a5-a259-475f-a87d-da53fa736a07"),
        "ppc64el": uuid.UUID("ee2b9983-21e8-4153-86d9-b6901a54d1ce"),
        "riscv32": uuid.UUID("cb1ee4e3-8cd0-4136-a0a4-aa61a32e8730"),
        "riscv64": uuid.UUID("8f1056be-9b05-47c4-81d6-be53128e5b54"),
        "s390": uuid.UUID("b663c618-e7bc-4d6d-90aa-11b756bb1797"),
        "s390x": uuid.UUID("31741cc4-1a2a-4111-a581-e00b447d2d06"),
        "tilegx": uuid.UUID("2fb4bf56-07fa-42da-8132-6b139f2026ae"),
        "x86_64": uuid.UUID("77ff5f63-e7b6-4633-acf4-1565b864c0e6"),
    },
    GPT_VOL_ARCH_USR_VERITY_SIG: {
        "alpha": uuid.UUID("5c6e1c76-076a-457a-a0fe-f3b4cd21ce6e"),
        "arc": uuid.UUID("94f9a9a1-9971-427a-a400-50cb297f0f35"),
        "arm": uuid.UUID("d7ff812f-37d1-4902-a810-d76ba57b975a"),
        "aarch64": uuid.UUID("c23ce4ff-44bd-4b00-b2d4-b41b3419e02a"),
        "i386": uuid.UUID("974a71c0-de41-43c3-be5d-5c5ccd1ad2c0"),
        "ia64": uuid.UUID("8de58bc2-2a43-460d-b14e-a76e4a17b47f"),
        "loongarch64": uuid.UUID("b024f315-d330-444c-8461-44bbde524e99"),
        "mips64el": uuid.UUID("f2c2c7ee-adcc-4351-b5c6-ee9816b66e16"),
        "mipsel": uuid.UUID("3e23ca0b-a4bc-4b4e-8087-5ab6a26aa8a9"),
        "parisc": uuid.UUID("450dd7d1-3224-45ec-9cf2-a43a346d71ee"),
        "ppc": uuid.UUID("7007891d-d371-4a80-86a4-5cb875b9302e"),
        "ppc64": uuid.UUID("0b888863-d7f8-4d9e-9766-239fce4d58af"),
        "ppc64el": uuid.UUID("c8bfbd1e-268e-4521-8bba-bf314c399557"),
        "riscv32": uuid.UUID("c3836a13-3137-45ba-b583-b16c50fe5eb4"),
        "riscv64": uuid.UUID("d2f9000a-7a18-453f-b5cd-4d32f77a7b32"),
        "s390": uuid.UUID("17440e4f-a8d0-467f-a46e-3912ae6ef2c5"),
        "s390x": uuid.UUID("3f324816-667b-46ae-86ee-9b0c0c6c11b4"),
        "tilegx": uuid.UUID("4ede75e2-6ccc-4cc8-b9c7-70334b087510"),
        "x86_64": uuid.UUID("e7bb33fb-06cf-4e81-8273-e543b413e2e2"),
    },
}


def gpt_part_uuid_for_volume(voltype, arch=None):
    """
    :param path: the volume type as a GPT_VOL constant
    :type path: str
    :param arch: the architecture of the target install, None to use host arch
    :type path: str

    :returns: the GPT partition type UUID or None
    :rtype: str

    Given a volume type defined by one of the GPT_VOL constants,
    determine if there is a well known GPT partition type UUID
    associated to allow the partition to be automatically mounted.

    The architecture must be canonicalized in accordance with the
    blivet.arch logic. Usually it can be omitted, to request use
    of the host architecture detected by blivet.arch.get_arch.
    """
    if voltype in _gpt_common_uuid:
        return _gpt_common_uuid[voltype]
    elif voltype not in _gpt_arch_uuid:
        raise GPTVolUUIDError("Unknown volume type")

    if arch is None:
        arch = get_arch()

    if arch not in _gpt_arch_uuid[voltype]:
        raise GPTVolUUIDError("Unknown architecture %s" % arch)

    return _gpt_arch_uuid[voltype][arch]


def gpt_part_uuid_for_mountpoint(path, arch=None):
    """
    :param path: the absolute path at which a filesystem is to be mounted
    :type path: str
    :param arch: the architecture of the target install, None to use host arch
    :type path: str

    :returns: the GPT partition type UUID or None
    :rtype: str

    Given an absolute path at which a filesystem is to be mounted,
    determine if there is a well known GPT partition type UUID
    associated to allow the partition to be automatically mounted.

    The architecture must be canonicalized in accordance with the
    blivet.arch logic. Usually it can be omitted, to request use
    of the host architecture detected by blivet.arch.get_arch.
    """
    mapping = {
        "/": GPT_VOL_ARCH_ROOT,
        "/usr": GPT_VOL_ARCH_USR,

        "/efi": GPT_VOL_ESP,
        "/boot/efi": GPT_VOL_ESP,
        "/boot": GPT_VOL_XBOOTLDR,
        "/home": GPT_VOL_HOME,
        "/var": GPT_VOL_VAR,
        "/srv": GPT_VOL_SRV,
        "/tmp": GPT_VOL_TMP,
    }

    if path not in mapping:
        return None

    return gpt_part_uuid_for_volume(mapping[path], arch)



#
# lvm.py
# lvm functions
#
# Copyright (C) 2009-2014  Red Hat, Inc.  All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author(s): Dave Lehman <dlehman@redhat.com>
#

import math
import os
import re

from collections import namedtuple
import itertools

import gi
gi.require_version("BlockDev", "2.0")

from gi.repository import BlockDev as blockdev

import logging
log = logging.getLogger("blivet")

from . import raid
from ..size import Size
from ..i18n import N_
from ..flags import flags
from ..static_data import lvs_info
from ..tasks import availability
from ..util import run_program

from contextlib import contextmanager

# some of lvm's defaults that we have no way to ask it for
LVM_PE_START = Size("1 MiB")
LVM_PE_SIZE = Size("4 MiB")

# thinp constants
LVM_THINP_MIN_METADATA_SIZE = Size("2 MiB")
LVM_THINP_MAX_METADATA_SIZE = Size("16 GiB")
LVM_THINP_MIN_CHUNK_SIZE = Size("64 KiB")
LVM_THINP_MAX_CHUNK_SIZE = Size("1 GiB")
LVM_THINP_ADDRESSABLE_CHUNK_SIZE = Size("17455015526400 B")  # 15.88 TiB

# cache constants
LVM_CACHE_MIN_METADATA_SIZE = Size("8 MiB")
LVM_CACHE_MAX_METADATA_SIZE = Size("16 GiB")
LVM_CACHE_DEFAULT_MODE = blockdev.LVMCacheMode.WRITETHROUGH

raid_levels = raid.RAIDLevels(["linear", "striped", "raid0", "raid1", "raid4", "raid5", "raid6", "raid10"])
raid_seg_types = list(itertools.chain.from_iterable([level.names for level in raid_levels if level.name != "linear"]))

ThPoolProfile = namedtuple("ThPoolProfile", ["name", "desc"])
KNOWN_THPOOL_PROFILES = (ThPoolProfile("thin-generic", N_("Generic")),
                         ThPoolProfile("thin-performance", N_("Performance")))

EXTERNAL_DEPENDENCIES = [availability.BLOCKDEV_LVM_PLUGIN]

LVMETAD_SOCKET_PATH = "/run/lvm/lvmetad.socket"

safe_name_characters = "0-9a-zA-Z._-"

if hasattr(blockdev.LVMTech, "DEVICES"):
    try:
        blockdev.lvm.is_tech_avail(blockdev.LVMTech.DEVICES, 0)  # pylint: disable=no-member
    except blockdev.LVMError:
        HAVE_LVMDEVICES = False
    else:
        HAVE_LVMDEVICES = True
else:
    HAVE_LVMDEVICES = False

# list of devices that LVM is allowed to use
# with LVM >= 2.0.13 we'll use this for the --devices option and when creating
# the /etc/lvm/devices/system.devices file
# with older versions of LVM we will use this for the --config based filtering
_lvm_devices = set()


def _set_global_config():
    """lvm command accepts lvm.conf type arguments preceded by --config. """

    device_string = ""

    if not HAVE_LVMDEVICES:
        # now explicitly "accept" all LVM devices
        for device in _lvm_devices:
            device_string += "\"a|%s$|\"," % device

        # now add all devices to the "reject" filter
        device_string += "\"r|.*|\""

        filter_string = "filter=[%s]" % device_string

        config_string = " devices { %s } " % filter_string
    else:
        config_string = " "

    if not flags.lvm_metadata_backup:
        config_string += "backup {backup=0 archive=0} "
    config_string += "log {level=7 file=/tmp/lvm.log syslog=0}"

    blockdev.lvm.set_global_config(config_string)


def _set_lvm_devices():
    if not HAVE_LVMDEVICES:
        return

    blockdev.lvm.set_devices_filter(list(_lvm_devices))


def needs_config_refresh(fn):
    if not availability.BLOCKDEV_LVM_PLUGIN.available:
        return lambda *args, **kwargs: None

    def fn_with_refresh(*args, **kwargs):
        ret = fn(*args, **kwargs)
        _set_global_config()
        _set_lvm_devices()
        return ret

    return fn_with_refresh


@needs_config_refresh
def lvm_devices_add(path):
    """ Add a device (PV) to the list of devices LVM is allowed to use """
    log.debug("lvm filter: device %s added to the list of allowed devices", path)
    _lvm_devices.add(path)


@needs_config_refresh
def lvm_devices_remove(path):
    """ Remove a device (PV) to the list of devices LVM is allowed to use """
    log.debug("lvm filter: device %s removed from the list of allowed devices", path)
    try:
        _lvm_devices.remove(path)
    except KeyError:
        log.debug("%s wasn't in the devices list", path)
        return


@needs_config_refresh
def lvm_devices_reset():
    log.debug("lvm filter: clearing the lvm devices list")
    _lvm_devices.clear()


def lvm_devices_copy():
    return _lvm_devices.copy()


@needs_config_refresh
def lvm_devices_restore(devices):
    log.debug("lvm filter: restoring the lvm devices list to %s", ", ".join(list(devices)))
    _lvm_devices = devices


@contextmanager
def empty_lvm_devices():
    devices = lvm_devices_copy()
    lvm_devices_reset()

    yield

    lvm_devices_restore(devices)


def determine_parent_lv(internal_lv, lvs, lv_info):
    """Try to determine which of the lvs is the parent of the internal_lv

    :param internal_lv: the internal LV to determine parent LV from
    :type internal_lv: :class:`~.devices.lvm.LMVInternalLogicalVolumeDevice`
    :param lvs: LVs searched for a potential parent LV
    :type lvs: :class:`~.devices.lvm.LMVLogicalVolumeDevice`
    :param lv_info: all available information about LVs
    :type lv_info: dict

    """
    for lv in lvs:
        # parent LVs has to be part of the same VG
        if lv.vg.name != internal_lv.vg.name:
            continue

        # skip the internal_lv itself
        if internal_lv.lvname == lv.lvname:
            continue

        info = lv_info.get(lv.name)
        if info is None:
            # internal LVs look like "vg_name-[int_lv_name]" in lv_info so let's
            # try that too
            info = lv_info.get("%s-%s" % (lv.vg.name, lv.display_lvname))
        if info:
            # cache pools are internal LVs of cached LVs
            pool_name = info.pool_lv
            if pool_name and pool_name.strip("[]") == internal_lv.lvname:
                return lv

            # pools have internal data and metadata LVs
            data_lv_name = info.data_lv
            if data_lv_name and data_lv_name.strip("[]") == internal_lv.lvname:
                return lv
            metadata_lv_name = info.metadata_lv
            if metadata_lv_name and metadata_lv_name.strip("[]") == internal_lv.lvname:
                return lv

        # try name matching
        # check if the lv's name is the name of the internal LV without the suffix
        # e.g. 'pool' and 'pool_tmeta'
        if re.match(lv.lvname + internal_lv.name_suffix + "$", internal_lv.lvname):
            return lv

    return None


def lvmetad_socket_exists():
    return os.path.exists(LVMETAD_SOCKET_PATH)


def ensure_lv_is_writable(vg_name, lv_name):
    lv_info = lvs_info.cache.get("%s-%s" % (vg_name, lv_name))
    if lv_info is None:
        return

    if lv_info.attr[1] == 'w':
        return

    try:
        rc = run_program(['lvchange', '-prw', "%s/%s" % (vg_name, lv_name)])
    except OSError:
        rc = -1

    return rc == 0


def is_lvm_name_valid(name):
    # No . or ..
    if name == '.' or name == '..':
        return False

    # Check that all characters are in the allowed set and that the name
    # does not start with a -
    if not re.match('^[a-zA-Z0-9+_.][a-zA-Z0-9+_.-]*$', name):
        return False

    # According to the LVM developers, vgname + lvname is limited to 126 characters
    # minus the number of hyphens, and possibly minus up to another 8 characters
    # in some unspecified set of situations. Instead of figuring all of that out,
    # no one gets a vg or lv name longer than, let's say, 55.
    if len(name) > 55:
        return False

    return True


def recommend_thpool_chunk_size(thpool_size):
    # calculation of the recommended chunk size by LVM is so complicated that we
    # can't really replicate it, but we know that 64 KiB chunk size gives us
    # upper limit of ~15.88 TiB so we will just add 64 KiB to the chunk size
    # for every ~15.88 TiB of thinpool data size
    return min(math.ceil(thpool_size / LVM_THINP_ADDRESSABLE_CHUNK_SIZE) * LVM_THINP_MIN_CHUNK_SIZE,
               LVM_THINP_MAX_CHUNK_SIZE)


def is_valid_cache_md_size(md_size):
    return md_size >= LVM_CACHE_MIN_METADATA_SIZE and md_size <= LVM_CACHE_MAX_METADATA_SIZE



#
# iscsi.py - iscsi class
#
# Copyright (C) 2005, 2006  IBM, Inc.  All rights reserved.
# Copyright (C) 2006  Red Hat, Inc.  All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from . import errors
from . import udev
from . import util
from .flags import flags
from .i18n import _
from . import safe_dbus
import os
import re
import shutil
import time
import itertools
from collections import namedtuple

import gi
gi.require_version("GLib", "2.0")
from gi.repository import GLib

import logging
log = logging.getLogger("blivet")

# Note that stage2 copies all files under /sbin to /usr/sbin
ISCSID = ""
INITIATOR_FILE = "/etc/iscsi/initiatorname.iscsi"

ISCSI_MODULES = ['cxgb3i', 'bnx2i', 'be2iscsi']


STORAGED_SERVICE = "org.freedesktop.UDisks2"
STORAGED_PATH = "/org/freedesktop/UDisks2"
STORAGED_MANAGER_PATH = "/org/freedesktop/UDisks2/Manager"
MANAGER_IFACE = "org.freedesktop.UDisks2.Manager"
INITIATOR_IFACE = MANAGER_IFACE + ".ISCSI.Initiator"
SESSION_IFACE = STORAGED_SERVICE + ".ISCSI.Session"


def has_iscsi():
    global ISCSID

    if not os.access("/sys/module/iscsi_tcp", os.X_OK):
        return False

    if not ISCSID:
        location = shutil.which("iscsid")
        if not location:
            return False
        ISCSID = location
        log.info("ISCSID is %s", ISCSID)

    return True


TargetInfo = namedtuple("TargetInfo", ["ipaddr", "port"])


class NodeInfo(object):
    """Simple representation of node information."""
    def __init__(self, name, tpgt, address, port, iface):
        self.name = name
        self.tpgt = tpgt
        self.address = address
        self.port = port
        self.iface = iface
        # These get set by log_into_node, but *NOT* _login
        self.username = None
        self.password = None
        self.r_username = None
        self.r_password = None

    @property
    def conn_info(self):
        """The 5-tuple of connection info (no auth info). This form
        is useful for interacting with storaged.
        """
        return (self.name, self.tpgt, self.address, self.port, self.iface)


class LoginInfo(object):
    def __init__(self, node, logged_in):
        self.node = node
        self.logged_in = logged_in


def _to_node_infos(variant):
    """Transforms an 'a(sisis)' GLib.Variant into a list of NodeInfo objects"""
    return [NodeInfo(*info) for info in variant]


class iSCSIDependencyGuard(util.DependencyGuard):
    error_msg = "storaged iSCSI functionality not available"

    def _check_avail(self):
        try:
            if not safe_dbus.check_object_available(STORAGED_SERVICE, STORAGED_MANAGER_PATH, MANAGER_IFACE):
                return False
            # storaged is modular and we need to make sure it has the iSCSI module
            # loaded (this also autostarts storaged if it isn't running already)
            safe_dbus.call_sync(STORAGED_SERVICE, STORAGED_MANAGER_PATH, MANAGER_IFACE,
                                "EnableModules", GLib.Variant("(b)", (True,)))
        except safe_dbus.DBusCallError:
            return False
        return safe_dbus.check_object_available(STORAGED_SERVICE, STORAGED_MANAGER_PATH, INITIATOR_IFACE)


storaged_iscsi_required = iSCSIDependencyGuard()


class iSCSI(object):
    """ iSCSI utility class.

        This class will automatically discover and login to iBFT (or
        other firmware) configured iscsi devices when the startup() method
        gets called. It can also be used to manually configure iscsi devices
        through the add_target() method.

        As this class needs to make sure certain things like starting iscsid
        and logging in to firmware discovered disks only happens once
        and as it keeps a global list of all iSCSI devices it is implemented as
        a Singleton.

        .. warning::
            Since this is a singleton class, calling deepcopy() on the instance
            just returns ``self`` with no copy being created.
    """

    def __init__(self):
        # Dictionary mapping discovered TargetInfo data to lists of LoginInfo
        # data.
        self.discovered_targets = {}
        # This list contains nodes discovered through iBFT (or other firmware)
        self.ibft_nodes = []
        self._initiator = ""
        self.started = False
        self.ifaces = {}

        self.__connection = None

        if flags.ibft:
            try:
                initiatorname = self._call_initiator_method("GetFirmwareInitiatorName")[0]
                self._initiator = initiatorname
            except Exception as e:  # pylint: disable=broad-except
                log.info("failed to get initiator name from iscsi firmware: %s", str(e))

    # So that users can write iscsi() to get the singleton instance
    def __call__(self):
        return self

    def __deepcopy__(self, memo_dict):
        # pylint: disable=unused-argument
        return self

    @property
    @storaged_iscsi_required(critical=False, eval_mode=util.EvalMode.onetime)
    def available(self):
        return True

    @property
    def _connection(self):
        if not self.__connection:
            self.__connection = safe_dbus.get_new_system_connection()

        return self.__connection

    @storaged_iscsi_required(critical=True, eval_mode=util.EvalMode.onetime)
    def _call_initiator_method(self, method, args=None):
        """Class a method of the ISCSI.Initiator DBus object

        :param str method: name of the method to call
        :param params: arguments to pass to the method
        :type params: GLib.Variant

        """
        return safe_dbus.call_sync(STORAGED_SERVICE, STORAGED_MANAGER_PATH,
                                   INITIATOR_IFACE, method, args,
                                   connection=self._connection)

    @property
    def initiator_set(self):
        """True if initiator is set at our level."""
        return self._initiator != ""

    @property
    @storaged_iscsi_required(critical=False, eval_mode=util.EvalMode.onetime)
    def initiator(self):
        if self._initiator != "":
            return self._initiator

        # udisks returns initiatorname as a NULL terminated bytearray
        raw_initiator = bytes(self._call_initiator_method("GetInitiatorNameRaw")[0][:-1])
        return raw_initiator.decode("utf-8", errors="replace")

    @initiator.setter
    @storaged_iscsi_required(critical=True, eval_mode=util.EvalMode.onetime)
    def initiator(self, val):
        if self.initiator_set and val != self._initiator:
            raise ValueError(_("Unable to change iSCSI initiator name once set"))
        if len(val) == 0:
            raise ValueError(_("Must provide an iSCSI initiator name"))

        log.info("Setting up iSCSI initiator name %s", self.initiator)
        args = GLib.Variant("(sa{sv})", (val, None))
        self._call_initiator_method("SetInitiatorName", args)
        self._initiator = val

    def active_nodes(self, target=None):
        """Nodes logged in to"""
        if target:
            return [info.node for info in self.discovered_targets.get(target, [])
                    if info.logged_in]
        else:
            return [info.node for info in itertools.chain(*list(self.discovered_targets.values()))
                    if info.logged_in] + self.ibft_nodes

    @property
    def mode(self):
        if not self.active_nodes():
            return "none"
        if self.ifaces:
            return "bind"
        else:
            return "default"

    def _mark_node_active(self, node, active=True):
        """Mark node as one logged in to

           Returns False if not found
        """
        for login_infos in self.discovered_targets.values():
            for info in login_infos:
                if info.node is node:
                    info.logged_in = active
                    return True
        return False

    def _login(self, node_info, extra=None):
        """Try to login to the iSCSI node

        :type node_info: :class:`NodeInfo`
        :param dict extra: extra configuration for the node (e.g. authentication info)
        :raises :class:`~.safe_dbus.DBusCallError`: if login fails

        """

        if extra is None:
            extra = dict()
        extra["node.startup"] = GLib.Variant("s", "automatic")

        args = GLib.Variant("(sisisa{sv})", node_info.conn_info + (extra,))
        self._call_initiator_method("Login", args)

    @storaged_iscsi_required(critical=False, eval_mode=util.EvalMode.onetime)
    def _get_active_sessions(self):
        try:
            objects = safe_dbus.call_sync(STORAGED_SERVICE,
                                          STORAGED_PATH,
                                          'org.freedesktop.DBus.ObjectManager',
                                          'GetManagedObjects',
                                          None)[0]
        except safe_dbus.DBusCallError as e:
            log.info("iscsi: Failed to get active sessions: %s", str(e))
            return []

        sessions = (obj for obj in objects.keys() if re.match(r'.*/iscsi/session[0-9]+$', obj))

        active = []
        for session in sessions:
            properties = objects[session][SESSION_IFACE]
            active.append(NodeInfo(properties["target_name"],
                                   properties["tpgt"],
                                   properties["persistent_address"],
                                   properties["persistent_port"],
                                   None))

        return active

    @storaged_iscsi_required(critical=False, eval_mode=util.EvalMode.onetime)
    def _start_ibft(self):
        if not flags.ibft:
            return

        args = GLib.Variant("(a{sv})", ([], ))
        try:
            found_nodes, _n_nodes = self._call_initiator_method("DiscoverFirmware", args)
        except safe_dbus.DBusCallError as e:
            log.info("iscsi: No IBFT info found: %s", str(e))
            # an exception here means there is no ibft firmware, just return
            return

        found_nodes = _to_node_infos(found_nodes)
        active_nodes = self._get_active_sessions()
        for node in found_nodes:
            if any(node.name == a.name and node.tpgt == a.tpgt and
                   node.address == a.address and node.port == a.port for a in active_nodes):
                log.info("iscsi IBFT: already logged in node %s at %s:%s through %s",
                         node.name, node.address, node.port, node.iface)
                self.ibft_nodes.append(node)
            try:
                self._login(node)
                log.info("iscsi IBFT: logged into %s at %s:%s through %s",
                         node.name, node.address, node.port, node.iface)
                self.ibft_nodes.append(node)
            except safe_dbus.DBusCallError as e:
                log.error("Could not log into ibft iscsi target %s: %s",
                          node.name, str(e))

        self.stabilize()

    def stabilize(self):
        # Wait for udev to create the devices for the just added disks

        # It is possible when we get here the events for the new devices
        # are not send yet, so sleep to make sure the events are fired
        time.sleep(2)
        udev.settle()

    def create_interfaces(self, ifaces):
        for iface in ifaces:
            iscsi_iface_name = "iface%d" % len(self.ifaces)
            # iscsiadm -m iface -I iface0 --op=new
            util.run_program(["iscsiadm", "-m", "iface",
                              "-I", iscsi_iface_name, "--op=new"])
            # iscsiadm -m iface -I iface0 --op=update -n iface.net_ifacename -v eth0
            util.run_program(["iscsiadm", "-m", "iface",
                              "-I", iscsi_iface_name, "--op=update",
                              "-n", "iface.net_ifacename", "-v", iface])

            self.ifaces[iscsi_iface_name] = iface
            log.debug("created_interface %s:%s", iscsi_iface_name, iface)

    def delete_interfaces(self):
        if not self.ifaces:
            return None
        for iscsi_iface_name in self.ifaces:
            # iscsiadm -m iface -I iface0 --op=delete
            util.run_program(["iscsiadm", "-m", "iface",
                              "-I", iscsi_iface_name, "--op=delete"])
        self.ifaces = {}

    def startup(self):
        if self.started:
            return

        if not has_iscsi():
            return

        # make sure that the file /etc/iscsi/initiatorname.iscsi exists
        util.run_program(["systemctl", "start", "iscsi-init.service"])

        if self._initiator == "":
            log.info("no initiator set")
            return

        for fulldir in (os.path.join("/var/lib/iscsi", d) for d in
                        ['ifaces', 'isns', 'nodes', 'send_targets', 'slp', 'static']):
            if not os.path.isdir(fulldir):
                os.makedirs(fulldir, 0o755)

        log.info("iSCSI startup")
        util.run_program(['modprobe', '-a'] + ISCSI_MODULES)
        # iscsiuio is needed by Broadcom offload cards (bnx2i). Currently
        # not present in iscsi-initiator-utils for Fedora.
        iscsiuio = shutil.which('iscsiuio')
        if iscsiuio:
            log.debug("iscsi: iscsiuio is at %s", iscsiuio)
            util.run_program([iscsiuio])
        else:
            log.info("iscsi: iscsiuio not found.")

        # run the daemon
        util.run_program([ISCSID])
        time.sleep(1)

        self._start_ibft()
        self.started = True

    def discover(self, ipaddr, port="3260", username=None, password=None,
                 r_username=None, r_password=None):
        """
        Discover iSCSI nodes on the target available for login.

        If we are logged in a node discovered for specified target
        do not do the discovery again as it can corrupt credentials
        stored for the node (set_auth and get_auth are using database
        in /var/lib/iscsi/nodes which is filled by discovery). Just
        return nodes obtained and stored in the first discovery
        instead.

        Returns list of nodes user can log in.
        """

        if not has_iscsi():
            raise errors.ISCSIError(_("iSCSI not available"))
        if self._initiator == "":
            raise ValueError(_("No initiator name set"))

        if self.active_nodes(TargetInfo(ipaddr, port)):
            log.debug("iSCSI: skipping discovery of %s:%s due to active nodes",
                      ipaddr, port)
        else:
            self.startup()
            auth_info = dict()
            if username:
                auth_info["username"] = GLib.Variant("s", username)
            if password:
                auth_info["password"] = GLib.Variant("s", password)
            if r_username:
                auth_info["reverse-username"] = GLib.Variant("s", r_username)
            if r_password:
                auth_info["reverse-password"] = GLib.Variant("s", r_password)

            args = GLib.Variant("(sqa{sv})", (ipaddr, int(port), auth_info))
            nodes, _n_nodes = self._call_initiator_method("DiscoverSendTargets", args)

            found_nodes = _to_node_infos(nodes)
            t_info = TargetInfo(ipaddr, port)
            self.discovered_targets[t_info] = []
            for node in found_nodes:
                self.discovered_targets[t_info].append(LoginInfo(node, False))
                log.debug("discovered iSCSI node: %s", node.name)

        # only return the nodes we are not logged into yet
        return [info.node for info in self.discovered_targets[TargetInfo(ipaddr, port)]
                if not info.logged_in]

    def log_into_node(self, node, username=None, password=None,
                      r_username=None, r_password=None):
        """
        :param node: node to log into
        :type node: :class:`NodeInfo`
        :param str username: username to use when logging in
        :param str password: password to use when logging in
        :param str r_username: r_username to use when logging in
        :param str r_password: r_password to use when logging in
        """

        rc = False  # assume failure
        msg = ""

        auth_info = dict()
        if username:
            auth_info["username"] = GLib.Variant("s", username)
        if password:
            auth_info["password"] = GLib.Variant("s", password)
        if r_username:
            auth_info["reverse-username"] = GLib.Variant("s", r_username)
        if r_password:
            auth_info["reverse-password"] = GLib.Variant("s", r_password)

        try:
            self._login(node, auth_info)
            rc = True
            log.info("iSCSI: logged into %s at %s:%s through %s",
                     node.name, node.address, node.port, node.iface)
            if not self._mark_node_active(node):
                log.error("iSCSI: node not found among discovered")
            if username:
                node.username = username
            if password:
                node.password = password
            if r_username:
                node.r_username = r_username
            if r_password:
                node.r_password = r_password
        except safe_dbus.DBusCallError as e:
            msg = str(e)
            log.warning("iSCSI: could not log into %s: %s", node.name, msg)

        return (rc, msg)

    def add_target(self, ipaddr, port="3260", user=None, pw=None,
                   user_in=None, pw_in=None, target=None, iface=None,
                   discover_user=None, discover_pw=None,
                   discover_user_in=None, discover_pw_in=None):
        """
        Connect to iSCSI server specified by IP address and port
        and add all targets found on the server and authenticate if necessary.
        If the target parameter is set, connect only to this target.

        NOTE: the iSCSI target can have two sets of different authentication
              credentials - one for discovery and one for logging into nodes

        :param str ipaddr: target IP address
        :param str port: target port
        :param user: CHAP username for node login
        :type user: str or NoneType
        :param pw: CHAP password for node login
        :type pw: str or NoneType
        :param user_in: reverse CHAP username for node login
        :type user: str or NoneType
        :param pw_in: reverse CHAP password for node login
        :type pw_in: str or NoneType
        :param target: only add this target (if present)
        :type target: str or NoneType
        :param iface: interface to use
        :type iface: str or NoneType
        :param discover_user: CHAP username for discovery
        :type discover_user: str or NoneType
        :param discover_pw: CHAP password for discovery
        :type discover_pw: str or NoneType
        :param discover_user_in: reverse CHAP username for discovery
        :type discover_user: str or NoneType
        :param discover_pw_in: reverse CHAP password for discovery
        :type discover_pw_in: str or NoneType
        """

        found = 0
        logged_in = 0

        found_nodes = self.discover(ipaddr, port, discover_user, discover_pw,
                                    discover_user_in, discover_pw_in)
        if found_nodes is None:
            raise errors.ISCSIError(_("No iSCSI nodes discovered"))

        for node in found_nodes:
            if target and target != node.name:
                log.debug("iscsi: skipping logging to iscsi node '%s'", node.name)
                continue
            if iface:
                node_net_iface = self.ifaces.get(node.iface, node.iface)
                if iface != node_net_iface:
                    log.debug("iscsi: skipping logging to iscsi node '%s' via %s",
                              node.name, node_net_iface)
                    continue

            found = found + 1

            (rc, _msg) = self.log_into_node(node, user, pw, user_in, pw_in)
            if rc:
                logged_in = logged_in + 1

        if found == 0:
            raise errors.ISCSIError(_("No new iSCSI nodes discovered"))

        if logged_in == 0:
            raise errors.ISCSIError(_("Could not log in to any of the discovered nodes"))

        self.stabilize()

    def write(self, root, storage=None):  # pylint: disable=unused-argument
        if not self.initiator_set:
            return

        # copy "db" files.  *sigh*
        if os.path.isdir(root + "/var/lib/iscsi"):
            shutil.rmtree(root + "/var/lib/iscsi")
        if os.path.isdir("/var/lib/iscsi"):
            shutil.copytree("/var/lib/iscsi", root + "/var/lib/iscsi",
                            symlinks=True)

        # copy the initiator file too
        if not os.path.isdir(root + "/etc/iscsi"):
            os.makedirs(root + "/etc/iscsi", 0o755)
        shutil.copyfile(INITIATOR_FILE, root + INITIATOR_FILE)

    def get_node(self, name, address, port, iface):
        for node in self.active_nodes():
            if node.name == name and node.address == address and \
               node.port == int(port) and node.iface == iface:
                return node

        return None

    def get_node_disks(self, node, storage):
        node_disks = []
        iscsi_disks = (d for d in storage.devices if d.type == "iscsi")
        for disk in iscsi_disks:
            if disk.node == node:
                node_disks.append(disk)

        return node_disks


# Create iscsi singleton
iscsi = iSCSI()
""" An instance of :class:`iSCSI` """

# vim:tw=78:ts=4:et:sw=4



# mounts.py
# Active mountpoints cache.
#
# Copyright (C) 2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Vojtech Trefny <vtrefny@redhat.com>
#
from collections import defaultdict
from .udev import resolve_devspec
from . import util
from .devicelibs import btrfs

import logging
log = logging.getLogger("blivet")

import os


class _MountinfoCache(object):

    """ Cache for info from /proc/self/mountinfo. Looks up the root of the
        mount within the filesystem using a pair of mountpoint, mount
        source as keys.

        This is a very simple helper class for MountsCache, and would
        have to be altered for general purpose use.

        Note that only info for fstype btrfs is stored within the cache,
        as MountsCache methods only require information for btrfs.
    """

    def __init__(self):
        self._cache = None

    def _get_cache(self):
        """ Reads lines in /proc/self/mountinfo and builds a table. """
        cache = {}

        with open("/proc/self/mountinfo") as mountinfos:
            for line in mountinfos:
                fields = line.split()
                separator_index = fields.index("-")
                fstype = fields[separator_index + 1]
                if not fstype.startswith("btrfs"):
                    continue

                root = fields[3]
                mountpoint = fields[4]
                # store the canonical device path
                devspec = resolve_devspec(fields[separator_index + 2],
                                          sysname=True)
                cache[(devspec, mountpoint)] = root

        return cache

    def get_root(self, devspec, mountpoint):
        """ Retrieves the root of the mount within the filesystem
            that corresponds to devspec and mountpoint.

            :param str devspec: device specification
            :param str mountpoint: mountpoint
            :rtype: str or NoneType
            :returns: the root of the mount within the filesystem, if available
        """
        if self._cache is None:
            self._cache = self._get_cache()

        # get the canonical device path
        devspec = resolve_devspec(devspec, sysname=True)
        return self._cache.get((devspec, mountpoint))


class MountsCache(object):

    """ Cache object for system mountpoints; checks /proc/mounts and
        /proc/self/mountinfo for up-to-date information.
    """

    def __init__(self):
        self.mounts_hash = 0
        self.mountpoints = defaultdict(list)

    def get_mountpoints(self, devspec, subvolspec=None):
        """ Get mountpoints for selected device

            :param devscpec: device specification, eg. "/dev/vda1"
            :type devspec: str
            :param subvolspec: btrfs subvolume specification, eg. ID or name
            :type subvolspec: object (may be NoneType)
            :returns: list of mountpoints (path)
            :rtype: list of str or empty list

            .. note::
                Devices can be mounted on multiple paths, and paths can have multiple
                devices mounted to them (hiding previous mounts). Callers should take this into account.
        """
        self._cache_check()

        if subvolspec is not None:
            subvolspec = str(subvolspec)

        # devspec == None means "get 'nodev' mount points"
        if devspec not in (None, "tmpfs"):
            if devspec.startswith("/dev"):
                # try to avoid using resolve_devspec if possible
                name = os.path.realpath(devspec).split("/")[-1]
                if (name, subvolspec) in self.mountpoints.keys():
                    return self.mountpoints[(name, subvolspec)]

            # use the canonical device path (if available)
            canon_devspec = resolve_devspec(devspec, sysname=True)
            if canon_devspec is not None:
                devspec = canon_devspec
            else:
                # udev doesn't know about the given device, it can hardly be
                # mounted
                return []

        return self.mountpoints[(devspec, subvolspec)]

    def is_mountpoint(self, path):
        """ Check to see if a path is already mounted

            :param str path: Path to check
        """
        self._cache_check()

        return any(path in p for p in self.mountpoints.values())

    def _get_active_mounts(self):
        """ Get information about mounted devices from /proc/mounts and
            /proc/self/mountinfo

            Refreshes self.mountpoints with current mountpoint information
        """
        self.mountpoints = defaultdict(list)
        mountinfo = _MountinfoCache()

        with open("/proc/mounts") as mounts:
            for line in mounts:
                try:
                    (devspec, mountpoint, fstype, _rest) = line.split(None, 3)
                except ValueError:
                    log.error("failed to parse /proc/mounts line: %s", line)
                    continue

                # use the canonical device path (if available)
                if devspec.startswith("/dev"):
                    devspec = resolve_devspec(devspec, sysname=True) or devspec

                if fstype == "btrfs":
                    root = mountinfo.get_root(devspec, mountpoint)
                    if root is not None:
                        subvolspec = root[1:] or str(btrfs.MAIN_VOLUME_ID)
                        self.mountpoints[(devspec, subvolspec)].append(mountpoint)
                    else:
                        log.error("failed to obtain subvolspec for btrfs device %s at mountpoint %s", devspec, mountpoint)
                else:
                    self.mountpoints[(devspec, None)].append(mountpoint)

    def _cache_check(self):
        """ Computes the SHA256 hash on /proc/mounts and updates the cache on change
        """

        sha256hash = util.sha256_file("/proc/mounts")

        if sha256hash != self.mounts_hash:
            self.mounts_hash = sha256hash
            self._get_active_mounts()


mounts_cache = MountsCache()



# events/handler.py
# Event handler mixin class.
#
# Copyright (C) 2015-2016  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU Lesser General Public License v.2, or (at your option) any later
# version. This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY expressed or implied, including the implied
# warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
# the GNU Lesser General Public License for more details.  You should have
# received a copy of the GNU Lesser General Public License along with this
# program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat trademarks
# that are incorporated in the source code or documentation are not subject
# to the GNU Lesser General Public License and may only be used or
# replicated with the express permission of Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#

from six import add_metaclass

from ..callbacks import callbacks
from ..errors import DeviceError, EventHandlingError
from ..devices import DM_MAJORS, MD_MAJORS
from .. import udev
from ..threads import SynchronizedMeta
from ..static_data import mpath_members

from .changes import data as event_data
from .manager import event_manager

import logging
log = logging.getLogger("blivet")
event_log = logging.getLogger("blivet.event")


@add_metaclass(SynchronizedMeta)
class EventHandlerMixin(object):
    def __init__(self):
        event_manager.handler_cb = self.handle_event

    def handle_event(self, event, notify_cb):
        """ Handle an event on a block device.

            :param :class:`~.event.Event` event: information about the event
            :param callable notify_cb: notification callback

            TODO: Rename all this stuff so it's explicit that it only handles uevents.
        """
        # delegate event to appropriate handler
        handlers = {"add": self._handle_add_event,
                    "change": self._handle_change_event,
                    "remove": self._handle_remove_event}

        handler = handlers.get(event.action)
        if handler is not None:
            handler(event)

        if notify_cb is not None:
            notify_cb(event=event, changes=event_data.changes)

    def _event_device_is_dm(self, event):
        """ Return True if event operand is a dm device.

            Since this may be run on add events it does not require the dm/
            subdirectory be present in the device's sysfs root, unlike
            udev.device_is_dm.

            XXX Should this replace udev.device_is_dm?
        """
        return udev.device_get_major(event.info) in DM_MAJORS

    def _event_device_is_md(self, event):
        """ Return True if event operand is a dm device.

            Since this may be run on add events it does not require the md/
            subdirectory be present in the device's sysfs root, unlike
            udev.device_is_md.

            XXX Should this replace udev.device_is_md?
        """
        return udev.device_get_major(event.info) in MD_MAJORS

    def _should_ignore_add_event(self, event):
        """ Return True if event is an add event that should be ignored.

            add events on md and dm devices should be ignored in general. When the
            operation is complete there will be a change event.
        """
        # udev.device_is_md and udev.device_is_dm are not sufficient for this purpose
        # because there are sometimes events when dm/ or md/ subdir does not exist.
        return (event.action == "add" and
                (self._event_device_is_dm(event) or self._event_device_is_md(event)))

    def _event_device_is_physical_disk(self, event):
        """ Return True if event device is a physical disk. """
        # udev.device_is_md and udev.device_is_dm are not sufficient for this purpose
        # because there are sometimes events when dm/ or md/ subdir does not exist.
        return (udev.device_is_disk(event.info) and
                not self._event_device_is_dm(event) and
                not self._event_device_is_md(event))

    def _handle_add_event(self, event):
        """ Handle an "add" event.

            Add events should correlate to activation rather than creation in
            most cases. When a device is added, there is generally a change
            event on the new device's parent(s). The obvious exception to this
            rule is the addition of a new disk, which has no parents.
        """
        # ignore add on dm and md devices
        if self._should_ignore_add_event(event):
            return

        # Try to look up the device. Do not look it up by sysfs path since in all
        # likelihood blivet thinks the device is not active and therefore has no
        # sysfs path.
        # XXX We need an existing device here, so we should also be checking destroy actions.
        device = self.get_device_by_name(udev.device_get_name(event.info), hidden=True)
        if device is None:
            device = self.get_device_by_uuid(event.info.get("UUID_SUB", udev.device_get_uuid(event.info)),
                                             hidden=True)
            if device is None and udev.device_is_dm_luks(event.info):
                # Special case for first-time decrypted LUKS devices since we do not add a
                # device for the decrypted/mapped device until it has been opened.
                self.handle_device(event.info)
                device = self.get_device_by_name(udev.device_get_name(event.info), hidden=True)

        # XXX Don't change anything (except simple attributes?) if actions are being executed.
        if device is None and self._event_device_is_physical_disk(event):
            log.info("disk %s was added", udev.device_get_name(event.info))
            mpath_members.update_cache(udev.device_get_devname(event.info))
            self.handle_device(event.info)
        elif device is not None and device.exists:
            log.info("device %s was activated", device.name)
            # device was activated from outside, so update the sysfs path
            sysfs_path = udev.device_get_sysfs_path(event.info)
            if device.sysfs_path != sysfs_path:
                old_sysfs_path = device.sysfs_path
                device.sysfs_path = sysfs_path
                callbacks.attribute_changed(device=device, attr="sysfs_path",
                                            old=old_sysfs_path, new=sysfs_path)

    def _handle_format_change(self, event, device):
        helper_class = self._get_format_helper(event.info, device)  # pylint: disable=no-member
        helper = helper_class(self, event.info, device=device)
        new_type = helper.type_spec
        old_type = device.format.type

        new_uuid = helper._get_kwargs().get("uuid")
        old_uuid = device.format.uuid

        # If the type has changed it has been reformatted. Easy.
        # If the type is unchanged and the UUID has changed, it could be a change to
        # the UUID or a reformat. Is it worthwhile to try to differentiate?
        log.debug("old_type=%-8s ; old_uuid=%-16s", old_type, old_uuid)
        log.debug("new_type=%-8s ; new_uuid=%-16s", new_type, new_uuid)
        if new_type == old_type and new_uuid == old_uuid and not self.actions.processing:
            helper.update()
            return

        if self.actions.processing:
            return

        self.cancel_disk_actions(device.disks)

        # The device was reformatted, but we can't blindly remove all children since
        # it could have been a member of a still-intact container.
        if getattr(device.format, "container_uuid", None):
            container = device.children[0]
            if len(container.parents) > 1:
                try:
                    container.parents.remove(device)
                except DeviceError as e:
                    log.error("error removing member %s from container %s: %s",
                              device.name, container.name, str(e))
                    raise EventHandlingError("reformatted container member")

                callbacks.parent_removed(device=container, parent=device)

        self.recursive_remove(device, actions=False, remove_device=False)

        helper.run()

    def _handle_change_event(self, event):
        """ Handle a "change" event.

            This could be any number of things, including activation of md/dm
            devices.
        """
        name = udev.device_get_name(event.info)
        if (name.startswith("dm-") or name.startswith("md")) and name == event.info.sys_name:
            log.debug("ignoring event on virtual device %s with no symbolic name", name)
            return

        if self._event_device_is_md(event) or self._event_device_is_dm(event):
            self._handle_add_event(event)

        # Try to look up the device in the devicetree.
        # We're expecting the device to already be active, so we try the lookup
        # by sysfs path first.
        device = self.get_device_by_sysfs_path(udev.device_get_sysfs_path(event.info))
        if device is None:
            # Lookup by sysfs path failed. Try looking it up by name.
            device = self.get_device_by_name(udev.device_get_name(event.info))

        if device is None:
            log.debug("failed to look up device")
            return

        if not device.exists:
            log.debug("device lookup returned a non-existent device")
            return

        log.info("device %s was changed", device.name)

        helper_class = self._get_device_helper(event.info)  # pylint: disable=no-member
        if helper_class is not None:
            helper = helper_class(self, event.info, device=device)
            helper.update()

        # check for metadata changes, but do not make changes to the devicetree
        # during action processing
        self._handle_format_change(event, device)

    def _handle_remove_event(self, event):
        """ Handle a "remove" event.

            Remove events correlate to deactivation or removal, but we will
            use change events on parent devices to detect removal.
        """
        # XXX We need an existing device here.
        device = self.get_device_by_sysfs_path(udev.device_get_sysfs_path(event.info))
        if device is None:
            device = self.get_device_by_name(udev.device_get_name(event.info))

        if device is None:
            return

        if self._event_device_is_physical_disk(event):
            log.info("disk %s was removed", device.name)
            self._remove_device(device)
        else:
            if device.sysfs_path:
                old_sysfs_path = device.sysfs_path
                log.info("device %s was deactivated", device.name)
                # device was deactivated from outside, so clear the sysfs path
                device.sysfs_path = ''
                callbacks.attribute_changed(device=device, attr="sysfs_path",
                                            old=old_sysfs_path, new='')



# mdraid.py
# Device format classes for anaconda's storage configuration module.
#
# Copyright (C) 2009  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Dave Lehman <dlehman@redhat.com>
#

import gi
gi.require_version("BlockDev", "2.0")

from gi.repository import BlockDev as blockdev

from ..storage_log import log_method_call
from parted import PARTITION_RAID
from . import DeviceFormat, register_device_format
from ..i18n import N_
from ..tasks import availability

import logging
log = logging.getLogger("blivet")


class MDRaidMember(DeviceFormat):

    """ An mdraid member disk. """
    _type = "mdmember"
    _name = N_("software RAID")
    _udev_types = ["linux_raid_member", "ddf_raid_member", "isw_raid_member"]
    parted_flag = PARTITION_RAID
    _formattable = True                 # can be formatted
    _supported = True                   # is supported
    _linux_native = True                 # for clearpart
    _packages = ["mdadm"]               # required packages
    _ks_mountpoint = "raid."
    _plugin = availability.BLOCKDEV_MDRAID_PLUGIN

    def __init__(self, **kwargs):
        """
            :keyword device: path to block device node
            :keyword uuid: this member device's uuid
            :keyword exists: whether this is an existing format
            :type exists: bool
            :keyword md_uuid: the uuid of the array this device belongs to

            .. note::

                The 'device' kwarg is required for existing formats.
        """
        log_method_call(self, **kwargs)
        DeviceFormat.__init__(self, **kwargs)
        self.md_uuid = kwargs.get("md_uuid")

        self.biosraid = kwargs.get("biosraid")

    def __repr__(self):
        s = DeviceFormat.__repr__(self)
        s += ("  md_uuid = %(md_uuid)s  biosraid = %(biosraid)s" %
              {"md_uuid": self.md_uuid, "biosraid": self.biosraid})
        return s

    @property
    def dict(self):
        d = super(MDRaidMember, self).dict
        d.update({"md_uuid": self.md_uuid, "biosraid": self.biosraid})
        return d

    @property
    def formattable(self):
        return super(MDRaidMember, self).formattable and self._plugin.available

    @property
    def supported(self):
        return super(MDRaidMember, self).supported and self._plugin.available

    def _destroy(self, **kwargs):
        blockdev.md.destroy(self.device)

    @property
    def destroyable(self):
        return self._plugin.available

    @property
    def status(self):
        # XXX hack -- we don't have a nice way to see if the array is active
        return False

    @property
    def hidden(self):
        return super(MDRaidMember, self).hidden or self.biosraid

    @property
    def container_uuid(self):
        return self.md_uuid

    @container_uuid.setter
    def container_uuid(self, uuid):
        self.md_uuid = uuid


register_device_format(MDRaidMember)



# fsminsize.py
# Filesystem size gathering classes.
#
# Copyright (C) 2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Anne Mulhern <amulhern@redhat.com>

import abc

from six import add_metaclass

from ..errors import FSError
from .. import util
from ..size import Size

from . import availability
from . import fstask
from . import task


@add_metaclass(abc.ABCMeta)
class FSMinSize(task.BasicApplication, fstask.FSTask):

    """ An abstract class that represents min size information extraction. """

    description = "minimum filesystem size"

    options = abc.abstractproperty(doc="Options for use with app.")

    def _resize_command(self):
        return [str(self.ext)] + self.options + [self.fs.device]

    def _get_resize_info(self):
        """ Get info from fsresize program.

            :rtype: str
            :returns: output returned by fsresize program
        """
        error_msg = None
        try:
            (rc, out) = util.run_program_and_capture_output(self._resize_command())
            if rc:
                error_msg = "failed to gather info from resize program: %d" % rc
        except OSError as e:
            error_msg = "failed to gather info from resize program: %s" % e

        if error_msg:
            raise FSError(error_msg)
        return out

    @abc.abstractmethod
    def do_task(self):  # pylint: disable=arguments-differ
        """ Returns the minimum size for this filesystem object.

            :rtype: :class:`~.size.Size`
            :returns: the minimum size
            :raises FSError: if filesystem can not be obtained
        """
        raise NotImplementedError()


class Ext2FSMinSize(FSMinSize):

    ext = availability.RESIZE2FS_APP
    options = ["-P"]

    @property
    def depends_on(self):
        return [self.fs._info]

    def _extract_block_size(self):
        """ Extract block size from filesystem info.

            :returns: block size of fileystem or None
            :rtype: :class:`~.size.Size` or NoneType
        """
        if self.fs._current_info is None:
            return None

        block_size = None
        for line in (l.strip() for l in self.fs._current_info.splitlines() if l.startswith("Block size:")):
            try:
                block_size = int(line.split(" ")[-1])
                break
            except ValueError:
                continue

        return Size(block_size) if block_size else None

    def _extract_num_blocks(self, info):
        """ Extract the number of blocks from the resizefs info.

           :returns: the number of blocks or None
           :rtype: int or NoneType
        """
        num_blocks = None
        for line in info.splitlines():
            (text, _sep, value) = line.partition(":")
            if "minimum size of the filesystem" not in text:
                continue

            try:
                num_blocks = int(value.strip())
                break
            except ValueError:
                break

        return num_blocks

    def do_task(self):  # pylint: disable=arguments-differ
        error_msgs = self.availability_errors
        if error_msgs:
            raise FSError("\n".join(error_msgs))

        block_size = self._extract_block_size()
        if block_size is None:
            raise FSError("failed to get block size for %s filesystem on %s" % (self.fs.mount_type, self.fs.device))

        resize_info = self._get_resize_info()
        num_blocks = self._extract_num_blocks(resize_info)
        if num_blocks is None:
            raise FSError("failed to get minimum block number for %s filesystem on %s" % (self.fs.mount_type, self.fs.device))

        return block_size * num_blocks


class NTFSMinSize(FSMinSize):

    ext = availability.NTFSRESIZE_APP
    options = ["-m"]

    def _extract_min_size(self, info):
        """ Extract the minimum size from the resizefs info.

            :param str info: info obtained from resizefs prog
            :rtype: :class:`~.size.Size` or NoneType
            :returns: the minimum size, or None
        """
        min_size = None
        for line in info.splitlines():
            (text, _sep, value) = line.partition(":")
            if "Minsize" not in text:
                continue

            try:
                min_size = Size("%d MB" % int(value.strip()))
            except ValueError:
                break

        return min_size

    def do_task(self):  # pylint: disable=arguments-differ
        error_msgs = self.availability_errors
        if error_msgs:
            raise FSError("\n".join(error_msgs))

        resize_info = self._get_resize_info()
        min_size = self._extract_min_size(resize_info)
        if min_size is None:
            raise FSError("Unable to discover minimum size of filesystem on %s" % self.fs.device)
        return min_size


class UnimplementedFSMinSize(fstask.UnimplementedFSTask):
    pass



# threads.py
# Utilities related to multithreading.
#
# Copyright (C) 2014,2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU Lesser General Public License v.2, or (at your option) any later
# version. This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY expressed or implied, including the implied
# warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
# the GNU Lesser General Public License for more details.  You should have
# received a copy of the GNU Lesser General Public License along with this
# program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat trademarks
# that are incorporated in the source code or documentation are not subject
# to the GNU Lesser General Public License and may only be used or
# replicated with the express permission of Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#

import threading
from types import FunctionType
from abc import ABCMeta
from six import raise_from, wraps, PY3
import functools

from .errors import ThreadError
from .flags import flags

blivet_lock = threading.RLock(verbose=flags.debug_threads)


def _is_main_thread():
    if PY3:
        return threading.current_thread() == threading.main_thread()
    else:
        return threading.currentThread().name == "MainThread"  # pylint: disable=deprecated-method


def exclusive(m):
    """ Run a callable while holding the global lock. """
    @wraps(m, set(functools.WRAPPER_ASSIGNMENTS) & set(dir(m)))
    def run_with_lock(*args, **kwargs):
        with blivet_lock:
            if _is_main_thread():
                exn_info = get_thread_exception()
                if exn_info[1]:
                    clear_thread_exception()
                    raise_from(ThreadError("raising queued exception"), exn_info[1])

            return m(*args, **kwargs)

    return run_with_lock


class SynchronizedMeta(type):
    """ Metaclass that wraps all methods with the exclusive decorator.

        To prevent specific methods from being wrapped, add the method name(s)
        to a class attribute called _unsynchronized_methods (list of str).
    """
    def __new__(cls, name, bases, dct):
        new_dct = {}

        for n in dct:
            obj = dct[n]
            # Do not decorate class or static methods.
            if n in dct.get('_unsynchronized_methods', []):
                pass
            elif isinstance(obj, FunctionType):
                obj = exclusive(obj)
            elif isinstance(obj, property):
                obj = property(fget=exclusive(obj.__get__),
                               fset=exclusive(obj.__set__),
                               fdel=exclusive(obj.__delattr__),
                               doc=obj.__doc__)

            new_dct[n] = obj

        return super(SynchronizedMeta, cls).__new__(cls, name, bases, new_dct)


class SynchronizedABCMeta(SynchronizedMeta, ABCMeta):
    pass


#
# Facilities for storing/retrieving information about an unhandled exception in a thread.
#
_exception_thread = None
_thread_exception = None


def save_thread_exception(thread, exc_info):
    global _exception_thread
    global _thread_exception

    if _exception_thread is None or _thread_exception is None:
        return

    _exception_thread = thread
    _thread_exception = exc_info


def clear_thread_exception():
    global _exception_thread
    global _thread_exception

    _exception_thread = None
    _thread_exception = None


def get_thread_exception():
    return (_exception_thread, _thread_exception)



# devices/disk.py
# Classes to represent various types of disk-like devices.
#
# Copyright (C) 2009-2014  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#

import gi
gi.require_version("BlockDev", "2.0")
gi.require_version("GLib", "2.0")

from gi.repository import BlockDev as blockdev
from gi.repository import GLib

import os
from collections import namedtuple

from .. import errors
from .. import util
from ..devicelibs import disk as disklib
from ..flags import flags
from ..storage_log import log_method_call
from .. import udev
from ..size import Size
from ..tasks import availability

from ..fcoe import fcoe

import logging
log = logging.getLogger("blivet")

from .lib import Tags
from .storage import StorageDevice
from .network import NetworkStorageDevice
from .dm import DMDevice


class DiskDevice(StorageDevice):

    """ A local/generic disk.

        This is not the only kind of device that is treated as a disk. More
        useful than checking isinstance(device, DiskDevice) is checking
        device.is_disk.
    """
    _type = "disk"
    _partitionable = True
    _is_disk = True

    def __init__(self, name, fmt=None,
                 size=None, major=None, minor=None, sysfs_path='',
                 parents=None, serial=None, vendor="", model="", bus="", wwn=None,
                 uuid=None, exists=True):
        """
            :param name: the device name (generally a device node's basename)
            :type name: str
            :keyword size: the device's size
            :type size: :class:`~.size.Size`
            :keyword parents: a list of parent devices
            :type parents: list of :class:`StorageDevice`
            :keyword fmt: this device's formatting
            :type fmt: :class:`~.formats.DeviceFormat` or a subclass of it
            :keyword uuid: universally unique identifier (device -- not fs)
            :type uuid: str
            :keyword sysfs_path: sysfs device path
            :type sysfs_path: str
            :keyword removable: whether or not this is a removable device
            :type removable: bool
            :keyword serial: the ID_SERIAL_RAW, ID_SERIAL or ID_SERIAL_SHORT for
                             this device (which one is available)
            :type serial: str
            :keyword vendor: the manufacturer of this Device
            :type vendor: str
            :keyword model: manufacturer's device model string
            :type model: str
            :keyword bus: the interconnect this device uses
            :type bus: str
            :keyword str wwn: the disk's WWN

            DiskDevices always exist.
        """
        StorageDevice.__init__(self, name, fmt=fmt, size=size,
                               major=major, minor=minor, exists=exists,
                               sysfs_path=sysfs_path, parents=parents,
                               serial=serial, model=model,
                               vendor=vendor, bus=bus, uuid=uuid)

        self.wwn = wwn or None

        try:
            ssd = int(util.get_sysfs_attr(self.sysfs_path, "queue/rotational")) == 0
        except TypeError:  # get_sysfs_attr returns None from all error paths
            ssd = False

        self.tags.add(Tags.local)
        if ssd:
            self.tags.add(Tags.ssd)
        if bus == "usb":
            self.tags.add(Tags.usb)
        if self.removable:
            self.tags.add(Tags.removable)

    def _clear_local_tags(self):
        local_tags = set([Tags.local, Tags.ssd, Tags.usb, Tags.removable])
        self.tags = self.tags.difference(local_tags)

    def __repr__(self):
        s = StorageDevice.__repr__(self)
        s += ("  removable = %(removable)s  wwn = %(wwn)s" % {"removable": self.removable,
                                                              "wwn": self.wwn})
        return s

    @property
    def media_present(self):
        if flags.testing:
            return True

        # Some drivers (cpqarray <blegh>) make block device nodes for
        # controllers with no disks attached and then report a 0 size,
        # treat this as no media present
        return self.exists and self.current_size > Size(0)

    @property
    def description(self):
        return " ".join(s for s in (self.vendor, self.model, self.wwn) if s)

    def _pre_destroy(self):
        """ Destroy the device. """
        log_method_call(self, self.name, status=self.status)
        if not self.media_present:
            raise errors.DeviceError("cannot destroy disk with no media")

        StorageDevice._pre_destroy(self)

    @property
    def _volume(self):
        return disklib.volumes.get(self.path)

    @property
    def raid_system(self):
        return self._volume.system if self._volume is not None else None

    @property
    def raid_level(self):
        return self._volume.raid_type if self._volume is not None else None

    @property
    def raid_stripe_size(self):
        return self._volume.raid_stripe_size if self._volume is not None else None

    @property
    def raid_disk_count(self):
        return self._volume.raid_disk_count if self._volume is not None else None


class DiskFile(DiskDevice):

    """ This is a file that we will pretend is a disk.

        This is intended only for testing purposes. The benefit of this class
        is that you can instantiate a disk-like device with a working disklabel
        class as a non-root user. It is not known how the system will behave if
        partitions are committed to one of these disks.
    """
    _dev_dir = ""

    def __init__(self, name, fmt=None,
                 size=None, major=None, minor=None, sysfs_path='',
                 parents=None, serial=None, vendor="", model="", bus="",
                 exists=True):
        """
            :param str name: the full path to the backing regular file
            :keyword :class:`~.formats.DeviceFormat` fmt: the device's format
        """
        _name = os.path.basename(name)
        self._dev_dir = os.path.dirname(name)

        super(DiskFile, self).__init__(_name, fmt=fmt, size=size,
                                       major=major, minor=minor, sysfs_path=sysfs_path,
                                       parents=parents, serial=serial, vendor=vendor,
                                       model=model, bus=bus, exists=exists)

    #
    # Regular files do not have sysfs entries.
    #
    @property
    def sysfs_path(self):
        return ""

    @sysfs_path.setter
    def sysfs_path(self, value):
        pass

    def update_sysfs_path(self):
        pass

    def read_current_size(self):
        size = Size(0)
        if self.exists and os.path.exists(self.path):
            st = os.stat(self.path)
            size = Size(st.st_size)

        return size


class MultipathDevice(DMDevice):

    """ A multipath device """
    _type = "dm-multipath"
    _packages = ["device-mapper-multipath"]
    _partitionable = True
    _is_disk = True
    _external_dependencies = [availability.MULTIPATH_APP]

    def __init__(self, name, fmt=None, size=None, wwn=None,
                 parents=None, sysfs_path=''):
        """
            :param name: the device name (generally a device node's basename)
            :type name: str
            :keyword size: the device's size
            :type size: :class:`~.size.Size`
            :keyword parents: a list of parent devices
            :type parents: list of :class:`StorageDevice`
            :keyword fmt: this device's formatting
            :type fmt: :class:`~.formats.DeviceFormat` or a subclass of it
            :keyword sysfs_path: sysfs device path
            :type sysfs_path: str
            :keyword str wwn: the device's WWN

            MultipathDevices always exist. Blivet cannot create or destroy
            them.
        """
        DMDevice.__init__(self, name, fmt=fmt, size=size,
                          parents=parents, sysfs_path=sysfs_path,
                          exists=True)
        self.wwn = wwn or None

    @property
    def model(self):
        if not self.parents:
            return ""
        return self.parents[0].model

    @property
    def vendor(self):
        if not self.parents:
            return ""
        return self.parents[0].vendor

    @property
    def description(self):
        return "WWID %s" % self.wwn

    def add_parent(self, parent):
        """ Add a parent device to the mpath. """
        log_method_call(self, self.name, status=self.status)
        if self.status:
            self.teardown()
            self.parents.append(parent)
            self.setup()
        else:
            self.parents.append(parent)

    def _add_parent(self, parent):
        super(MultipathDevice, self)._add_parent(parent)
        if Tags.remote not in self.tags and Tags.remote in parent.tags:
            self.tags.add(Tags.remote)

    def _remove_parent(self, parent):
        super(MultipathDevice, self)._remove_parent(parent)
        if Tags.remote in self.tags and Tags.remote in parent.tags and \
           not any(p for p in self.parents if Tags.remote in p.tags and p != parent):
            self.tags.remove(Tags.remote)

    def _setup(self, orig=False):
        """ Open, or set up, a device. """
        log_method_call(self, self.name, orig=orig, status=self.status,
                        controllable=self.controllable)
        udev.settle()
        rc = util.run_program(["multipath", self.name])
        if rc:
            raise errors.MPathError("multipath activation failed for '%s'" %
                                    self.name, hardware_fault=True)

    def _post_setup(self):
        StorageDevice._post_setup(self)
        self.setup_partitions()
        udev.settle()


class iScsiDiskDevice(DiskDevice, NetworkStorageDevice):

    """ An iSCSI disk. """
    _type = "iscsi"
    _packages = ["iscsi-initiator-utils", "dracut-network"]

    def __init__(self, device, **kwargs):
        """
            :param name: the device name (generally a device node's basename)
            :type name: str
            :keyword exists: does this device exist?
            :type exists: bool
            :keyword size: the device's size
            :type size: :class:`~.size.Size`
            :keyword parents: a list of parent devices
            :type parents: list of :class:`StorageDevice`
            :keyword format: this device's formatting
            :type format: :class:`~.formats.DeviceFormat` or a subclass of it
            :keyword str wwn: the disk's WWN
            :keyword target: the name of the iscsi target
            :type target: str
            :keyword lun: lun of the target
            :type node: str
            :keyword iface: name of network interface to use for operation
            :type iface: str
            :keyword initiator: initiator name
            :type initiator: str
            :keyword offload: a partial offload device (qla4xxx)
            :type: bool
            :keyword address: ip address of the target
            :type: str
            :keyword port: port of the target
            :type: str
        """
        # Backward compatibility attributes - to be removed
        self.node = kwargs.pop("node")
        self.ibft = kwargs.pop("ibft")
        self.nic = kwargs.pop("nic")

        self.initiator = kwargs.pop("initiator")
        self.offload = kwargs.pop("offload")
        name = kwargs.pop("name")
        self.target = kwargs.pop("target")
        try:
            self.lun = int(kwargs.pop("lun"))
        except TypeError as e:
            log.warning("Failed to set lun attribute of iscsi disk: %s", e)
            self.lun = None

        self.address = kwargs.pop("address")
        self.port = kwargs.pop("port")
        self.iface = kwargs.pop("iface")
        self.id_path = kwargs.pop("id_path")
        DiskDevice.__init__(self, device, **kwargs)
        NetworkStorageDevice.__init__(self, host_address=self.address, nic=self.iface)
        log.debug("created new iscsi disk %s from target: %s lun: %s portal: %s:%s interface: %s partial offload: %s)",
                  name, self.target, self.lun, self.address, self.port, self.iface, self.offload)

        self._clear_local_tags()

    def dracut_setup_args(self):
        if self.ibft:
            return set(["rd.iscsi.firmware"])

        # qla4xxx partial offload
        if self.node is None:
            return set()

        address = self.node.address
        # surround ipv6 addresses with []
        if ":" in address:
            address = "[%s]" % address

        netroot = "netroot=iscsi:"
        if self.node.username and self.node.password:
            netroot += "%s:%s" % (self.node.username, self.node.password)
            if self.node.r_username and self.node.r_password:
                netroot += ":%s:%s" % (self.node.r_username,
                                       self.node.r_password)

        iface_spec = ""
        if self.nic != "default":
            iface_spec = ":%s:%s" % (self.node.iface, self.nic)
        netroot += "@%s::%d%s::%s" % (address,
                                      self.node.port,
                                      iface_spec,
                                      self.node.name)

        initiator = "rd.iscsi.initiator=%s" % self.initiator

        return set([netroot, initiator])


class FcoeDiskDevice(DiskDevice, NetworkStorageDevice):

    """ An FCoE disk. """
    _type = "fcoe"
    _packages = ["fcoe-utils", "dracut-network"]

    def __init__(self, device, **kwargs):
        """
            :param name: the device name (generally a device node's basename)
            :type name: str
            :keyword exists: does this device exist?
            :type exists: bool
            :keyword size: the device's size
            :type size: :class:`~.size.Size`
            :keyword parents: a list of parent devices
            :type parents: list of :class:`StorageDevice`
            :keyword format: this device's formatting
            :type format: :class:`~.formats.DeviceFormat` or a subclass of it
            :keyword str wwn: the disk's WWN
            :keyword nic: name of NIC to use
            :keyword identifier: ???
        """
        self.nic = kwargs.pop("nic")
        self.identifier = kwargs.pop("identifier")
        self.id_path = kwargs.pop("id_path")
        DiskDevice.__init__(self, device, **kwargs)
        NetworkStorageDevice.__init__(self, nic=self.nic)
        log.debug("created new fcoe disk %s (%s) @ %s",
                  device, self.identifier, self.nic)

        self._clear_local_tags()

    def dracut_setup_args(self):
        dcb = True

        for nic, dcb, _auto_vlan in fcoe().nics:
            if nic == self.nic:
                break
        else:
            return set()

        if dcb:
            dcb_opt = "dcb"
        else:
            dcb_opt = "nodcb"

        if self.nic in fcoe().added_nics:
            return set(["fcoe=%s:%s" % (self.nic, dcb_opt)])
        else:
            return set(["fcoe=edd:%s" % dcb_opt])


class ZFCPDiskDevice(DiskDevice):

    """ A mainframe ZFCP disk. """
    _type = "zfcp"

    def __init__(self, device, **kwargs):
        """
            :param name: the device name (generally a device node's basename)
            :type name: str
            :keyword exists: does this device exist?
            :type exists: bool
            :keyword size: the device's size
            :type size: :class:`~.size.Size`
            :keyword parents: a list of parent devices
            :type parents: list of :class:`StorageDevice`
            :keyword format: this device's formatting
            :type format: :class:`~.formats.DeviceFormat` or a subclass of it
            :keyword str wwn: the disk's WWN
            :keyword hba_id: ???
            :keyword wwpn: ???
            :keyword fcp_lun: ???
        """
        self.hba_id = kwargs.pop("hba_id")
        self.wwpn = kwargs.pop("wwpn")
        self.fcp_lun = kwargs.pop("fcp_lun")
        DiskDevice.__init__(self, device, **kwargs)
        self._clear_local_tags()
        self.tags.add(Tags.remote)

    def __repr__(self):
        s = DiskDevice.__repr__(self)
        s += ("  hba_id = %(hba_id)s  wwpn = %(wwpn)s  fcp_lun = %(fcp_lun)s" %
              {"hba_id": self.hba_id,
               "wwpn": self.wwpn,
               "fcp_lun": self.fcp_lun})
        return s

    @property
    def description(self):
        return "FCP device %(device)s with WWPN %(wwpn)s and LUN %(lun)s" \
               % {'device': self.hba_id,
                  'wwpn': self.wwpn,
                  'lun': self.fcp_lun}

    def dracut_setup_args(self):
        from ..zfcp import has_auto_lun_scan

        # zFCP auto LUN scan needs only the device ID
        if has_auto_lun_scan(self.hba_id):
            dracut_args = set(["rd.zfcp=%s" % self.hba_id])
        else:
            dracut_args = set(["rd.zfcp=%s,%s,%s" % (self.hba_id, self.wwpn, self.fcp_lun,)])

        return dracut_args


class DASDDevice(DiskDevice):

    """ A mainframe DASD. """
    _type = "dasd"

    def __init__(self, device, **kwargs):
        """
            :param name: the device name (generally a device node's basename)
            :type name: str
            :keyword exists: does this device exist?
            :type exists: bool
            :keyword size: the device's size
            :type size: :class:`~.size.Size`
            :keyword parents: a list of parent devices
            :type parents: list of :class:`StorageDevice`
            :keyword format: this device's formatting
            :type format: :class:`~.formats.DeviceFormat` or a subclass of it
            :keyword str wwn: the disk's WWN
            :keyword busid: bus ID
            :keyword opts: options
            :type opts: dict with option name keys and option value values
        """
        self.busid = kwargs.pop('busid')
        self.opts = kwargs.pop('opts')
        DiskDevice.__init__(self, device, **kwargs)

    @property
    def description(self):
        return "DASD device %s" % self.busid

    def get_opts(self):
        return ["%s=%s" % (k, v) for k, v in self.opts.items() if v == '1']

    def dracut_setup_args(self):
        conf = "/etc/dasd.conf"
        line = None
        if os.path.isfile(conf):
            f = open(conf)
            # grab the first line that starts with our bus_id
            for l in f.readlines():
                if l.startswith(self.busid):
                    line = l.rstrip()
                    break

            f.close()

        # See if we got a line.  If not, grab our get_opts
        if not line:
            line = self.busid
            for devopt in self.get_opts():
                line += " %s" % devopt

        # Create a translation mapping from dasd.conf format to module format
        translate = {'use_diag': 'diag',
                     'readonly': 'ro',
                     'erplog': 'erplog',
                     'failfast': 'failfast'}

        # this is a really awkward way of determining if the
        # feature found is actually desired (1, not 0), plus
        # translating that feature into the actual kernel module
        # value
        opts = []
        parts = line.split()
        for chunk in parts[1:]:
            try:
                feat, val = chunk.split('=')
                if int(val):
                    opts.append(translate[feat])
            except (ValueError, KeyError):
                # If we don't know what the feature is (feat not in translate
                # or if we get a val that doesn't cleanly convert to an int
                # we can't do anything with it.
                log.warning("failed to parse dasd feature %s", chunk)

        if opts:
            return set(["rd.dasd=%s(%s)" % (self.busid,
                                            ":".join(opts))])
        else:
            return set(["rd.dasd=%s" % self.busid])


class NVDIMMNamespaceDevice(DiskDevice):

    """ Non-volatile memory namespace """
    _type = "nvdimm"
    _packages = ["ndctl"]

    def __init__(self, device, **kwargs):
        """
            :param name: the device name (generally a device node's basename)
            :type name: str
            :keyword exists: does this device exist?
            :type exists: bool
            :keyword size: the device's size
            :type size: :class:`~.size.Size`
            :keyword parents: a list of parent devices
            :type parents: list of :class:`StorageDevice`
            :keyword format: this device's formatting
            :type format: :class:`~.formats.DeviceFormat` or a subclass of it
            :keyword mode: mode of the namespace
            :type mode: str
            :keyword devname: name of the namespace (e.g. 'namespace0.0')
            :type devname: str
            :keyword sector_size: sector size of the namespace in sector mode
            :type sector_size: str
        """
        self.mode = kwargs.pop("mode")
        self.devname = kwargs.pop("devname")
        self.id_path = kwargs.pop("id_path")
        self._sector_size = kwargs.pop("sector_size")

        DiskDevice.__init__(self, device, **kwargs)

        self._clear_local_tags()
        self.tags.add(Tags.local)
        self.tags.add(Tags.nvdimm)

    def __repr__(self):
        s = DiskDevice.__repr__(self)
        s += ("  mode = %(mode)s  devname = %(devname)s" %
              {"mode": self.mode,
               "devname": self.devname})
        if self.sector_size:
            s += ("  sector size = %(sector_size)s" % {"sector_size": self.sector_size})
        return s

    @property
    def description(self):
        return "NVDIMM namespace %(devname)s in %(mode)s mode exported as %(path)s" \
               % {'devname': self.devname,
                  'mode': self.mode,
                  'path': self.path}

    @property
    def sector_size(self):
        return self._sector_size


NVMeController = namedtuple("NVMeController", ["name", "serial", "nvme_ver", "id", "subsysnqn",
                                               "transport", "transport_address"])


class NVMeNamespaceDevice(DiskDevice):

    """ NVMe namespace """
    _type = "nvme"
    _packages = ["nvme-cli"]

    def __init__(self, device, **kwargs):
        """
            :param name: the device name (generally a device node's basename)
            :type name: str
            :keyword exists: does this device exist?
            :type exists: bool
            :keyword size: the device's size
            :type size: :class:`~.size.Size`
            :keyword parents: a list of parent devices
            :type parents: list of :class:`StorageDevice`
            :keyword format: this device's formatting
            :type format: :class:`~.formats.DeviceFormat` or a subclass of it
            :keyword nsid: namespace ID
            :type nsid: int
        """
        self.nsid = kwargs.pop("nsid", 0)
        self.eui64 = kwargs.pop("eui64", "")
        self.nguid = kwargs.pop("nguid", "")

        DiskDevice.__init__(self, device, **kwargs)

        self._clear_local_tags()
        self.tags.add(Tags.local)
        self.tags.add(Tags.nvme)

        self._controllers = None

    @property
    def controllers(self):
        if self._controllers is not None:
            return self._controllers

        self._controllers = []
        if not hasattr(blockdev.Plugin, "NVME"):
            # the nvme plugin is not generally available
            log.debug("Failed to get controllers for %s: libblockdev NVME plugin is not available", self.name)
            return self._controllers

        try:
            controllers = blockdev.nvme_find_ctrls_for_ns(self.sysfs_path)
        except GLib.GError as err:
            log.debug("Failed to get controllers for %s: %s", self.name, str(err))
            return self._controllers

        for controller in controllers:
            try:
                cpath = util.get_path_by_sysfs_path(controller, "char")
            except RuntimeError as err:
                log.debug("Failed to find controller %s: %s", controller, str(err))
                continue
            try:
                cinfo = blockdev.nvme_get_controller_info(cpath)
            except GLib.GError as err:
                log.debug("Failed to get controller info for %s: %s", cpath, str(err))
                continue
            ctrans = util.get_sysfs_attr(controller, "transport")
            ctaddr = util.get_sysfs_attr(controller, "address")
            self._controllers.append(NVMeController(name=os.path.basename(cpath),
                                                    serial=cinfo.serial_number,
                                                    nvme_ver=cinfo.nvme_ver,
                                                    id=cinfo.ctrl_id,
                                                    subsysnqn=cinfo.subsysnqn,
                                                    transport=ctrans,
                                                    transport_address=ctaddr))

        return self._controllers


class NVMeFabricsNamespaceDevice(NVMeNamespaceDevice, NetworkStorageDevice):

    """ NVMe fabrics namespace """
    _type = "nvme-fabrics"
    _packages = ["nvme-cli"]

    def __init__(self, device, **kwargs):
        """
            :param name: the device name (generally a device node's basename)
            :type name: str
            :keyword exists: does this device exist?
            :type exists: bool
            :keyword size: the device's size
            :type size: :class:`~.size.Size`
            :keyword parents: a list of parent devices
            :type parents: list of :class:`StorageDevice`
            :keyword format: this device's formatting
            :type format: :class:`~.formats.DeviceFormat` or a subclass of it
        """
        NVMeNamespaceDevice.__init__(self, device, **kwargs)
        NetworkStorageDevice.__init__(self)

        self._clear_local_tags()
        self.tags.add(Tags.remote)
        self.tags.add(Tags.nvme)



# populator.py
# Backend code for populating a DeviceTree.
#
# Copyright (C) 2009-2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU Lesser General Public License v.2, or (at your option) any later
# version. This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY expressed or implied, including the implied
# warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
# the GNU Lesser General Public License for more details.  You should have
# received a copy of the GNU Lesser General Public License along with this
# program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat trademarks
# that are incorporated in the source code or documentation are not subject
# to the GNU Lesser General Public License and may only be used or
# replicated with the express permission of Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#

from ... import udev
from ...devices import OpticalDevice
from ...storage_log import log_method_call
from .devicepopulator import DevicePopulator


class OpticalDevicePopulator(DevicePopulator):
    @classmethod
    def match(cls, data):
        return udev.device_is_cdrom(data)

    def run(self):
        log_method_call(self)
        # XXX should this be RemovableDevice instead?
        #
        # Looks like if it has ID_INSTANCE=0:1 we can ignore it.
        device = OpticalDevice(udev.device_get_name(self.data),
                               major=udev.device_get_major(self.data),
                               minor=udev.device_get_minor(self.data),
                               sysfs_path=udev.device_get_sysfs_path(self.data),
                               vendor=udev.device_get_vendor(self.data),
                               model=udev.device_get_model(self.data))
        self._devicetree._add_device(device)
        return device



#
# nvdimm.py - nvdimm class
#
# Copyright (C) 2018  Red Hat, Inc.  All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import gi
gi.require_version("BlockDev", "2.0")
gi.require_version("GLib", "2.0")
from gi.repository import BlockDev
from gi.repository import GLib

from .. import util

import logging
log = logging.getLogger("blivet")


class NVDIMMDependencyGuard(util.DependencyGuard):
    error_msg = "libblockdev NVDIMM functionality not available"

    def _check_avail(self):
        try:
            BlockDev.nvdimm_is_tech_avail(BlockDev.NVDIMMTech.NVDIMM_TECH_NAMESPACE,
                                          BlockDev.NVDIMMTechMode.RECONFIGURE |
                                          BlockDev.NVDIMMTechMode.QUERY |
                                          BlockDev.NVDIMMTechMode.ACTIVATE_DEACTIVATE)
        except GLib.GError:
            return False
        return True


blockdev_nvdimm_required = NVDIMMDependencyGuard()


class NVDIMM(object):
    """ NVDIMM utility class.

        .. warning::
            Since this is a singleton class, calling deepcopy() on the instance
            just returns ``self`` with no copy being created.
    """

    def __init__(self):
        self._namespaces = None

    # So that users can write nvdimm() to get the singleton instance
    def __call__(self):
        return self

    def __deepcopy__(self, memo_dict):
        # pylint: disable=unused-argument
        return self

    @property
    def namespaces(self):
        """ Dict of all NVDIMM namespaces, including dax and disabled namespaces
        """
        if not self._namespaces:
            self.update_namespaces_info()

        return self._namespaces

    @blockdev_nvdimm_required(critical=True, eval_mode=util.EvalMode.onetime)
    def update_namespaces_info(self):
        """ Update information about the namespaces
        """
        namespaces = BlockDev.nvdimm_list_namespaces(idle=True)

        self._namespaces = dict((namespace.dev, namespace) for namespace in namespaces)

    def get_namespace_info(self, device):
        """ Get namespace information for a device
            :param str device: device name (e.g. 'pmem0') or path
        """
        for info in self.namespaces.values():
            if info.blockdev == device or \
               (device.startswith("/dev/") and info.blockdev == device[5:]):
                return info

    @blockdev_nvdimm_required(critical=True, eval_mode=util.EvalMode.onetime)
    def enable_namespace(self, namespace):
        """ Enable a namespace
            :param str namespace: devname of the namespace (e.g. 'namespace0.0')
        """

        if namespace not in self.namespaces.keys():
            raise ValueError("Namespace '%s' doesn't exist." % namespace)

        BlockDev.nvdimm_namespace_enable(namespace)

        # and update our namespaces info "cache"
        self.update_namespaces_info()

    @blockdev_nvdimm_required(critical=True, eval_mode=util.EvalMode.onetime)
    def reconfigure_namespace(self, namespace, mode, **kwargs):
        """ Change mode of the namespace
            :param str namespace: devname of the namespace (e.g. 'namespace0.0')
            :param str mode: new mode of the namespace (one of 'sector', 'memory', 'dax')
            :keyword int sector_size: sector size when reconfiguring to the 'sector' mode
            :keyword str map_location: map location when reconfiguring to the 'memory'
                                       mode (one of 'mem', 'dev')

            .. note::
                This doesn't change state of the devicetree. It is necessary to
                run reset() or populate() to make these changes visible.
        """

        if namespace not in self.namespaces.keys():
            raise ValueError("Namespace '%s' doesn't exist." % namespace)

        info = self.namespaces[namespace]

        sector_size = kwargs.get("sector_size", None)
        map_location = kwargs.get("map_location", None)

        if sector_size and mode != "sector":
            raise ValueError("Sector size cannot be set for selected mode '%s'." % mode)

        if map_location and mode != "memory":
            raise ValueError("Map location cannot be set for selected mode '%s'." % mode)

        mode_t = BlockDev.nvdimm_namespace_get_mode_from_str(mode)

        if sector_size:
            extra = {"-l": str(sector_size)}
        elif map_location:
            extra = {"-M": map_location}
        else:
            extra = None

        BlockDev.nvdimm_namespace_reconfigure(namespace, mode_t, info.enabled, extra)

        # and update our namespaces info "cache"
        self.update_namespaces_info()


# Create nvdimm singleton
nvdimm = NVDIMM()
""" An instance of :class:`NVDIMM` """



#
# zfcp.py - mainframe zfcp configuration install data
#
# Copyright (C) 2001, 2002, 2003, 2004  Red Hat, Inc.  All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author(s): Karsten Hopp <karsten@redhat.com>
#

import os
import re
from abc import ABC, abstractmethod
import glob
from . import udev
from . import util
from .i18n import _
from .util import stringize, unicodeize

import gi
gi.require_version("BlockDev", "2.0")

from gi.repository import BlockDev as blockdev

import logging
log = logging.getLogger("blivet")


def logged_write_line_to_file(fn, value):
    f = open(fn, "w")
    log.debug("echo %s > %s", value, fn)
    f.write("%s\n" % (value))
    f.close()


zfcpsysfs = "/sys/bus/ccw/drivers/zfcp"
scsidevsysfs = "/sys/bus/scsi/devices"
zfcpconf = "/etc/zfcp.conf"


def _is_lun_scan_allowed():
    """Return True if automatic LUN scanning is enabled by the kernel."""

    allow_lun_scan = util.get_kernel_module_parameter("zfcp", "allow_lun_scan")
    return allow_lun_scan == "Y"


def _is_port_in_npiv_mode(device_id):
    """Return True if the device ID is configured in NPIV mode. See
    https://www.ibm.com/docs/en/linux-on-systems?topic=devices-use-npiv
    """

    port_in_npiv_mode = False
    port_type_path = "/sys/bus/ccw/devices/{}/host*/fc_host/host*/port_type".format(device_id)
    port_type_paths = glob.glob(port_type_path)
    try:
        for filename in port_type_paths:
            with open(filename) as f:
                port_type = f.read()
            if re.search(r"(^|\s)NPIV(\s|$)", port_type):
                port_in_npiv_mode = True
    except OSError as e:
        log.warning("Couldn't read the port_type attribute of the %s device: %s", device_id, str(e))
        port_in_npiv_mode = False

    return port_in_npiv_mode


def has_auto_lun_scan(device_id):
    """Return True if the given zFCP device ID is configured in NPIV (N_Port ID Virtualization)
    mode and zFCP auto LUN scan is not disabled.

    :returns: True or False
    """

    # LUN scanning disabled by the kernel module prevents using zFCP auto LUN scan
    if not _is_lun_scan_allowed():
        log.warning("Automatic LUN scanning is disabled by the zfcp kernel module.")
        return False

    # The port itself has to be configured in NPIV mode
    if not _is_port_in_npiv_mode(device_id):
        log.warning("The zFCP device %s is not configured in NPIV mode.", device_id)
        return False

    return True


class ZFCPDeviceBase(ABC):
    """An abstract base class for zFCP storage devices."""

    def __init__(self, devnum):
        self.devnum = blockdev.s390.sanitize_dev_input(devnum)
        if not self.devnum:
            raise ValueError(_("You have not specified a device number or the number is invalid"))

        self._device_online_path = os.path.join(zfcpsysfs, self.devnum, "online")

    # Force str and unicode types in case any of the properties are unicode
    def _to_string(self):
        return str(self.devnum)

    def __str__(self):
        return stringize(self._to_string())

    def __unicode__(self):
        return unicodeize(self._to_string())

    def _free_device(self):
        """Remove the device from the I/O ignore list to make it visible to the system.

        :raises: ValueError if the device cannot be removed from the I/O ignore list
        """

        if not os.path.exists(self._device_online_path):
            log.info("Freeing zFCP device %s", self.devnum)
            util.run_program(["zfcp_cio_free", "-d", self.devnum])

        if not os.path.exists(self._device_online_path):
            raise ValueError(_("zFCP device %s not found, not even in device ignore list.") %
                             (self.devnum,))

    def _set_zfcp_device_online(self):
        """Set the zFCP device online.

        :raises: ValueError if the device cannot be set online
        """

        try:
            with open(self._device_online_path) as f:
                devonline = f.readline().strip()
            if devonline != "1":
                logged_write_line_to_file(self._device_online_path, "1")
        except OSError as e:
            raise ValueError(_("Could not set zFCP device %(devnum)s "
                               "online (%(e)s).")
                             % {'devnum': self.devnum, 'e': e})

    def _set_zfcp_device_offline(self):
        """Set the zFCP device offline.

        :raises: ValueError if the device cannot be set offline
        """

        try:
            logged_write_line_to_file(self._device_online_path, "0")
        except OSError as e:
            raise ValueError(_("Could not set zFCP device %(devnum)s "
                               "offline (%(e)s).")
                             % {'devnum': self.devnum, 'e': e})

    @abstractmethod
    def _is_associated_with_fcp(self, fcphbasysfs, fcpwwpnsysfs, fcplunsysfs):
        """Decide if the provided FCP addressing corresponds to the path stored in the zFCP device.

        :returns: True or False
        """

    def online_device(self):
        """Initialize the device and make its storage block device(s) ready to use.

        :returns: True if success
        :raises: ValueError if the device cannot be initialized
        """

        self._free_device()
        self._set_zfcp_device_online()
        return True

    def offline_scsi_device(self):
        """Find SCSI devices associated to the zFCP device and remove them from the system."""

        # A list of existing SCSI devices in format Host:Bus:Target:Lun
        scsi_devices = [f for f in os.listdir(scsidevsysfs) if re.search(r'^[0-9]+:[0-9]+:[0-9]+:[0-9]+$', f)]

        scsi_device_found = False
        for scsidev in scsi_devices:
            fcpsysfs = os.path.join(scsidevsysfs, scsidev)

            with open(os.path.join(fcpsysfs, "hba_id")) as f:
                fcphbasysfs = f.readline().strip()
            with open(os.path.join(fcpsysfs, "wwpn")) as f:
                fcpwwpnsysfs = f.readline().strip()
            with open(os.path.join(fcpsysfs, "fcp_lun")) as f:
                fcplunsysfs = f.readline().strip()

            if self._is_associated_with_fcp(fcphbasysfs, fcpwwpnsysfs, fcplunsysfs):
                scsi_device_found = True
                scsidel = os.path.join(scsidevsysfs, scsidev, "delete")
                logged_write_line_to_file(scsidel, "1")
                udev.settle()

        if not scsi_device_found:
            log.warning("No scsi device found to delete for zfcp %s", self)


class ZFCPDeviceFullPath(ZFCPDeviceBase):
    """A class for zFCP devices where zFCP auto LUN scan is not available. Such
    devices have to be specified by a device number, WWPN and LUN.
    """

    def __init__(self, devnum, wwpn, fcplun):
        super().__init__(devnum)

        self.wwpn = blockdev.s390.zfcp_sanitize_wwpn_input(wwpn)
        if not self.wwpn:
            raise ValueError(_("You have not specified a worldwide port name or the name is invalid."))

        self.fcplun = blockdev.s390.zfcp_sanitize_lun_input(fcplun)
        if not self.fcplun:
            raise ValueError(_("You have not specified a FCP LUN or the number is invalid."))

    # Force str and unicode types in case any of the properties are unicode
    def _to_string(self):
        return "{} {} {}".format(self.devnum, self.wwpn, self.fcplun)

    def _is_associated_with_fcp(self, fcphbasysfs, fcpwwpnsysfs, fcplunsysfs):
        """Decide if the provided FCP addressing corresponds to the path stored in the zFCP device.

        :returns: True or False
        """

        return (fcphbasysfs == self.devnum and
                fcpwwpnsysfs == self.wwpn and
                fcplunsysfs == self.fcplun)

    def online_device(self):
        """Initialize the device and make its storage block device(s) ready to use.

        :returns: True if success
        :raises: ValueError if the device cannot be initialized
        """

        super().online_device()

        portadd = "%s/%s/port_add" % (zfcpsysfs, self.devnum)
        portdir = "%s/%s/%s" % (zfcpsysfs, self.devnum, self.wwpn)
        unitadd = "%s/unit_add" % (portdir)
        unitdir = "%s/%s" % (portdir, self.fcplun)
        failed = "%s/failed" % (unitdir)

        # Activating using devnum, WWPN, and LUN despite available zFCP auto LUN scan should still
        # be possible as this method was used as a workaround until the support for zFCP auto LUN
        # scan devices has been implemented. Just log a warning message and continue.
        if has_auto_lun_scan(self.devnum):
            log.warning("zFCP device %s in NPIV mode brought online. All LUNs will be activated "
                        "automatically although WWPN and LUN have been provided.", self.devnum)

        # create the sysfs directory for the WWPN/port
        if not os.path.exists(portdir):
            if os.path.exists(portadd):
                # older zfcp sysfs interface
                try:
                    logged_write_line_to_file(portadd, self.wwpn)
                    udev.settle()
                except OSError as e:
                    raise ValueError(_("Could not add WWPN %(wwpn)s to zFCP "
                                       "device %(devnum)s (%(e)s).")
                                     % {'wwpn': self.wwpn,
                                         'devnum': self.devnum,
                                         'e': e})
            else:
                # newer zfcp sysfs interface with auto port scan
                raise ValueError(_("WWPN %(wwpn)s not found at zFCP device "
                                   "%(devnum)s.") % {'wwpn': self.wwpn,
                                                     'devnum': self.devnum})
        else:
            if os.path.exists(portadd):
                # older zfcp sysfs interface
                log.info("WWPN %(wwpn)s at zFCP device %(devnum)s already "
                         "there.", {'wwpn': self.wwpn,
                                    'devnum': self.devnum})

        # create the sysfs directory for the LUN/unit
        if not os.path.exists(unitdir):
            try:
                logged_write_line_to_file(unitadd, self.fcplun)
                udev.settle()
            except OSError as e:
                raise ValueError(_("Could not add LUN %(fcplun)s to WWPN "
                                   "%(wwpn)s on zFCP device %(devnum)s "
                                   "(%(e)s).")
                                 % {'fcplun': self.fcplun, 'wwpn': self.wwpn,
                                     'devnum': self.devnum, 'e': e})
        else:
            raise ValueError(_("LUN %(fcplun)s at WWPN %(wwpn)s on zFCP "
                               "device %(devnum)s already configured.")
                             % {'fcplun': self.fcplun,
                                 'wwpn': self.wwpn,
                                 'devnum': self.devnum})

        # check the state of the LUN
        fail = "0"
        try:
            f = open(failed, "r")
            fail = f.readline().strip()
            f.close()
        except OSError as e:
            raise ValueError(_("Could not read failed attribute of LUN "
                               "%(fcplun)s at WWPN %(wwpn)s on zFCP device "
                               "%(devnum)s (%(e)s).")
                             % {'fcplun': self.fcplun,
                                 'wwpn': self.wwpn,
                                 'devnum': self.devnum,
                                 'e': e})
        if fail != "0":
            self.offline_device()
            raise ValueError(_("Failed LUN %(fcplun)s at WWPN %(wwpn)s on "
                               "zFCP device %(devnum)s removed again.")
                             % {'fcplun': self.fcplun,
                                 'wwpn': self.wwpn,
                                 'devnum': self.devnum})

        return True

    def offline_device(self):
        """Remove the zFCP device from the system."""

        portadd = "%s/%s/port_add" % (zfcpsysfs, self.devnum)
        portremove = "%s/%s/port_remove" % (zfcpsysfs, self.devnum)
        unitremove = "%s/%s/%s/unit_remove" % (zfcpsysfs, self.devnum, self.wwpn)
        portdir = "%s/%s/%s" % (zfcpsysfs, self.devnum, self.wwpn)
        devdir = "%s/%s" % (zfcpsysfs, self.devnum)

        try:
            self.offline_scsi_device()
        except OSError as e:
            raise ValueError(_("Could not correctly delete SCSI device of "
                               "zFCP %(devnum)s %(wwpn)s %(fcplun)s "
                               "(%(e)s).")
                             % {'devnum': self.devnum, 'wwpn': self.wwpn,
                                 'fcplun': self.fcplun, 'e': e})

        # remove the LUN
        try:
            logged_write_line_to_file(unitremove, self.fcplun)
        except OSError as e:
            raise ValueError(_("Could not remove LUN %(fcplun)s at WWPN "
                               "%(wwpn)s on zFCP device %(devnum)s "
                               "(%(e)s).")
                             % {'fcplun': self.fcplun, 'wwpn': self.wwpn,
                                 'devnum': self.devnum, 'e': e})

        # remove the WWPN only if there are no other LUNs attached
        if os.path.exists(portadd):
            # only try to remove ports with older zfcp sysfs interface
            for lun in os.listdir(portdir):
                if lun.startswith("0x") and \
                        os.path.isdir(os.path.join(portdir, lun)):
                    log.info("Not removing WWPN %s at zFCP device %s since port still has other LUNs, e.g. %s.",
                             self.wwpn, self.devnum, lun)
                    return True

            try:
                logged_write_line_to_file(portremove, self.wwpn)
            except OSError as e:
                raise ValueError(_("Could not remove WWPN %(wwpn)s on zFCP "
                                   "device %(devnum)s (%(e)s).")
                                 % {'wwpn': self.wwpn,
                                     'devnum': self.devnum, 'e': e})

        # check if there are other WWPNs existing for the zFCP device number
        if os.path.exists(portadd):
            # older zfcp sysfs interface
            for port in os.listdir(devdir):
                if port.startswith("0x") and \
                        os.path.isdir(os.path.join(devdir, port)):
                    log.info("Not setting zFCP device %s offline since it still has other ports, e.g. %s.",
                             self.devnum, port)
                    return True
        else:
            # newer zfcp sysfs interface with auto port scan
            luns = glob.glob("%s/0x????????????????/0x????????????????"
                             % (devdir,))
            if len(luns) != 0:
                log.info("Not setting zFCP device %s offline since it still has other LUNs, e.g. %s.",
                         self.devnum, luns[0])
                return True

        # no other WWPNs/LUNs exists for this device number, it's safe to bring it offline
        self._set_zfcp_device_offline()

        return True


class ZFCPDevice(ZFCPDeviceFullPath):
    """Class derived from ZFCPDeviceFullPath to reserve backward compatibility for applications
    using the ZFCPDevice class. ZFCPDeviceFullPath should be used instead in new code.
    """


class ZFCPDeviceAutoLunScan(ZFCPDeviceBase):
    """Class for zFCP devices configured in NPIV mode and zFCP auto LUN scan not disabled. Only
    a zFCP device number is needed for such devices.
    """

    def online_device(self):
        """Initialize the device and make its storage block device(s) ready to use.

        :returns: True if success
        :raises: ValueError if the device cannot be initialized
        """

        super().online_device()

        if not has_auto_lun_scan(self.devnum):
            raise ValueError(_("zFCP device %s cannot use auto LUN scan.") % self)

        return True

    def offline_device(self):
        """Remove the zFCP device from the system.

        :returns: True if success
        :raises: ValueError if the device cannot be brought offline
         """

        try:
            self.offline_scsi_device()
        except OSError as e:
            raise ValueError(_("Could not correctly delete SCSI device of "
                               "zFCP %(zfcpdev)s (%(e)s).")
                             % {'zfcpdev': self, 'e': e})

        self._set_zfcp_device_offline()

        return True

    def _is_associated_with_fcp(self, fcphbasysfs, _fcpwwpnsysfs, _fcplunsysfs):
        """Decide if the provided FCP addressing corresponds to the zFCP device.

        :returns: True or False
        """

        return fcphbasysfs == self.devnum


class zFCP:

    """ ZFCP utility class.

        This class will automatically online to ZFCP drives configured in
        /tmp/fcpconfig when the startup() method gets called. It can also be
        used to manually configure ZFCP devices through the add_fcp() method.

        As this class needs to make sure that /tmp/fcpconfig configured
        drives are only onlined once and as it keeps a global list of all ZFCP
        devices it is implemented as a Singleton.
    """

    def __init__(self):
        self.fcpdevs = set()
        self.has_read_config = False
        self.down = True

    # So that users can write zfcp() to get the singleton instance
    def __call__(self):
        return self

    def __deepcopy__(self, memo_dict):
        # pylint: disable=unused-argument
        return self

    def read_config(self):
        try:
            f = open(zfcpconf, "r")
        except OSError:
            log.info("no %s; not configuring zfcp", zfcpconf)
            return

        lines = [x.strip().lower() for x in f.readlines()]
        f.close()

        for line in lines:
            if line.startswith("#") or line == '':
                continue

            fields = line.split()

            # zFCP auto LUN scan available
            if len(fields) == 1:
                devnum = fields[0]
                wwpn = None
                fcplun = None
            elif len(fields) == 3:
                devnum = fields[0]
                wwpn = fields[1]
                fcplun = fields[2]
            elif len(fields) == 5:
                # support old syntax of:
                # devno scsiid wwpn scsilun fcplun
                devnum = fields[0]
                wwpn = fields[2]
                fcplun = fields[4]
            else:
                log.warning("Invalid line found in %s: %s", zfcpconf, line)
                continue

            try:
                self.add_fcp(devnum, wwpn, fcplun)
            except ValueError as e:
                log.warning("%s", str(e))

    def add_fcp(self, devnum, wwpn=None, fcplun=None):
        if wwpn and fcplun:
            d = ZFCPDeviceFullPath(devnum, wwpn, fcplun)
        else:
            d = ZFCPDeviceAutoLunScan(devnum)

        if d.online_device():
            self.fcpdevs.add(d)

    def shutdown(self):
        if self.down:
            return
        self.down = True
        if len(self.fcpdevs) == 0:
            return
        for d in self.fcpdevs:
            try:
                d.offline_device()
            except ValueError as e:
                log.warning("%s", str(e))

    def startup(self):
        if not self.down:
            return
        self.down = False
        if not self.has_read_config:
            self.read_config()
            self.has_read_config = True
            # read_config calls add_fcp which calls online_device already
            return

        if len(self.fcpdevs) == 0:
            return
        for d in self.fcpdevs:
            try:
                d.online_device()
            except ValueError as e:
                log.warning("%s", str(e))

    def write(self, root):
        if len(self.fcpdevs) == 0:
            return
        f = open(root + zfcpconf, "w")
        for d in self.fcpdevs:
            f.write("%s\n" % (d,))
        f.close()

        f = open(root + "/etc/modprobe.conf", "a")
        f.write("alias scsi_hostadapter zfcp\n")
        f.close()


# Create ZFCP singleton
zfcp = zFCP()

# vim:tw=78:ts=4:et:sw=4



from collections import namedtuple

from . import raid
from .. import errors
from .. import util
from ..size import Size

lsm = None

_HBA_PLUGIN_URIS = ("hpsa://", "megaraid://")
LSMInfo = namedtuple('HBAVolumeInfo', ['system', 'nodes', 'raid_type', 'raid_stripe_size', 'raid_disk_count'])
""" .. class:: LSMInfo

        .. attribute:: system (str): descriptive name of HBA unit
        .. attribute:: nodes (list[str]): list of device node paths for the volume
        .. attribute:: raid_type (:class:`~.devicelibs.raid.RAIDLevel` or None): RAID level
        .. attribute:: raid_stripe_size (:class:`~.size.Size` or None): stripe size
        .. attribute:: raid_disk_count (int or None): number of disks in the RAID set
"""
volumes = dict()
_raid_levels = dict()


class _LSMRAIDLevelStub(raid.RAIDLevel):
    def __init__(self, name):
        self._name = name

    @property
    def name(self):
        return self._name

    @property
    def names(self):
        return [self.name]

    @property
    def min_members(self):
        return 0

    def has_redundancy(self):
        return False

    def is_uniform(self):
        return False


class _LSMDependencyGuard(util.DependencyGuard):
    error_msg = "libstoragemgmt functionality not available"

    def _check_avail(self):
        global lsm
        if lsm is None:  # pylint: disable=used-before-assignment
            try:
                import lsm  # pylint: disable=redefined-outer-name
            except ImportError:
                lsm = None

        return lsm is not None


_lsm_required = _LSMDependencyGuard()


def _update_lsm_raid_levels():
    """ Build a mapping of lsm.RAID_TYPE->blivet.devicelibs.raid.RAIDLevel """
    global _raid_levels
    _raid_levels = dict()
    lsm_raid_levels = dict((k, v) for (k, v) in lsm.Volume.__dict__.items() if k.startswith("RAID_TYPE_"))
    for constant_name, value in lsm_raid_levels.items():
        name = constant_name[len("RAID_TYPE_"):]
        try:
            level = raid.get_raid_level(name)
        except errors.RaidError:
            level = _LSMRAIDLevelStub(name)

        _raid_levels[value] = level


def _get_lsm_raid_level(lsm_raid_type):
    """ Return a blivet.devicelibs.raid.RAIDLevel corresponding the lsm-reported RAID level."""
    return _raid_levels.get(lsm_raid_type, _raid_levels.get(lsm.Volume.RAID_TYPE_UNKNOWN))


@_lsm_required(critical=False, eval_mode=util.EvalMode.always)
def update_volume_info():
    """ Build a dict of namedtuples containing basic HBA RAID info w/ device path keys. """
    global volumes
    volumes = dict()
    _update_lsm_raid_levels()
    for uri in _HBA_PLUGIN_URIS:
        try:
            client = lsm.Client(uri)
        except lsm.LsmError:
            continue

        systems = dict((s.id, s) for s in client.systems())
        for vol in client.volumes():
            nodes = lsm.LocalDisk.vpd83_search(vol.vpd83)
            system = systems[vol.system_id]
            caps = client.capabilities(system)

            raid_level = None
            stripe_size = None
            disk_count = None
            if caps.supported(lsm.Capabilities.VOLUME_RAID_INFO):
                raid_info = client.volume_raid_info(vol)[:3]
                raid_level = _get_lsm_raid_level(raid_info[0])
                stripe_size = Size(raid_info[1])
                disk_count = raid_info[2]

            info = LSMInfo(system.name, nodes, raid_level, stripe_size, disk_count)
            volumes.update([(node, info) for node in nodes])
